// Bigram tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "bigram.h"
#include <limits>
using namespace std;

//---------------------------------------------------------------------------
bool Bigram::readData(istream &in, TagSet &tagset, Lexicon &lex, Endings &rules)
{
    if (tagset.readBinary(in) && rules.read(tagset, in) && lex.read(tagset, in))
    {
        return true;
    }
    else
    {
        lex.clear();
        rules.clear();
        tagset.clear();

        cerr << "Error(s) reading data file" << endl;
        return false;
    }
}

void Bigram::writeData(ostream &out, const TagSet &tagset,
                       const Lexicon &lex, const Endings &rules, bool textual)
{
    tagset.write(out, textual);
    
    // Only write rules if binary, since if textual they will be identical to
    // the input
    if (!textual)
    {
        rules.write(out);
    }
    lex.write(out, textual);
}

//-------------------------------------------------------------------------------
void Bigram::TagSpine::fill_node(Lexicon::LexEntry &entry, const Endings &rules)
{
    clear_node();

    // Fill in fields
    lex    = &entry;
    chosen = 0;

    // Deal with the special case of an unknown word
    if (entry.isUnknown())
    {
        // Make a local copy of the data
        local = true;
        tags  = new Lexicon::LexTags();
        entry.copyTags(*tags);
        
        // Adjust according to endings rules
        unsigned int ntags = rules.apply(token, *tags);
        if (ntags == 0)
        {
            // Lost everything, so restore the tags from the lexical entry.
            entry.copyTags(*tags);
        }
    }

    if (tags == NULL || tags->size() == 0)
    {
        // Share tags with the lexical entry.
        local = false;
        tags  = entry.getTags();
    }

    // Fill in algorithm specific data.
    filler();
}

//------------------------------------------------------------------------------
void Bigram::FBSpine::filler()
{
    ct = 0;
    data.clear();
    data.resize(tags->size());
}

void Bigram::FBSpine::clearer()
{
    ct = 0;
    data.clear();
}

void Bigram::ViterbiSpine::filler()
{
    data.clear();
    data.resize(tags->size());
}

void Bigram::ViterbiSpine::clearer()
{
    data.clear();
}

//---------------------------------------------------------------------------
void Bigram::Tagger::clear_chart()
{
    // Clear hypotheses from everything except final word
    if (chart.size() > 0)
    {
        TagSpine *last = chart.back();
        chart.pop_back();
        for (TagChart::iterator i = chart.begin(); i != chart.end(); ++i)
        {
            TagSpine *at = *i;
            at->clear_node();
            delete at;
        }
        chart.clear();
        chart.push_back(last);
    }
}

void Bigram::Tagger::destroy_chart()
{
    for (TagChart::iterator i = chart.begin(); i != chart.end(); ++i)
    {
        TagSpine *at = *i;
        at->clear_node();
        delete at;
    }
    chart.clear();
}

void Bigram::Tagger::new_chart(const Lexicon::LexEntry &anchor)
{
    destroy_chart();
    Endings dummy_rules;

    TagSpine *node = new_node();
    node->fill_node((Lexicon::LexEntry &) anchor, dummy_rules);
    chart.push_back(node);
}

void Bigram::Tagger::tagger
(const TokenReader *reader, const Lexicon &lex, const TagSet &tagset,
 const Endings &rules, istream &in, bool casehold, bool initial)
{
    bool done = false;
    new_chart(*lex.getAnchor());

    while (!done)
    {
        // Add a node for whatever we read, or an anchor if we read nothing
        TagSpine *node = new_node();
        chart.push_back(node);
        
        TaggedToken &token = node->getToken();
        string normal_form;

        // Read a token: we read as if it were tagged, even if it is not
        if (!reader->read(in, token))
        {
            // End of input: add final anchor
            token.anchor(tagset);
            done = true;
        }

        // "Normalise" even if we have the final anchor
        normal_form = token.getToken();
        token_handler(token);
        TagBase::normalise(normal_form, lex.getNumberTest(), true);

        // Locate the lexical entry, possibly changing the case
        Lexicon::LexEntry *lexeme = lex.find(normal_form) ;
        if (lexeme->isUnknown() && !casehold)
        {
            TagBase::lowercase(normal_form);
            lexeme = lex.find(normal_form);
        }

        // Invoke unknown handler
        if (lexeme->isUnknown())
        {
            unknown_handler(lex, token);
        }

        // Add the token to the chart, and check if we are ready to tag
        node->fill_node(*lexeme, rules);

        if (node->tagCount() == 1)
        {
            // Node has a single tag, so we are able to run the tagging algorithm
            tag();

            // Write the results.
            writer(initial);

            // Flush the chart for the next time round
            clear_chart();
        }
    }

    // Clear up anything that's left
    destroy_chart();
}

//---------------------------------------------------------------------------
// Forward-backward tagging (with numerical stabilisation)
// This is a slight variant on standard FB tagging, in that we assume the
// initial word has a single tag, and so we need not use a pi score for it.
void Bigram::FBTagger::tag()
{
    alpha_pass();
    beta_pass ();
    gamma_pass();
}

void Bigram::FBTagger::alpha_pass()
{
    TagChart::iterator i = chart.begin();
    FBSpine *from = (FBSpine *) *i;
    double ct;

    // Initialise scores on first node. This is equivalent to assigning the
    // lexical score, then doing the numerical stabilisation calculation.
    from->getData(0).alpha = 1;
    from->setStable(ct = from->getTag(0).score);
    from->setChosen(0);

    // Main body of pass
    for (i++; i != chart.end(); i++)
    {
        FBSpine *to = (FBSpine *) *i;

        double alpha_total = 0;
        unsigned int fc, tc; // from and to counters

        unsigned int from_ntags = from->tagCount();
        unsigned int to_ntags   = to->tagCount();

        for (tc = 0; tc < to_ntags; ++tc)
        {
            Lexicon::LexTag &toTag = to->getTag(tc);
            FBDatum &th = to->getData(tc);
            th.alpha = TagBase::tinyScore;
            

            for (fc = 0; fc < from_ntags ; ++fc)
            {
                FBDatum &fh = from->getData(fc);
                th.alpha += fh.alpha * from->getTag(fc).transScore(toTag);
            }

            th.alpha = th.alpha * toTag.score + TagBase::tinyScore;
            alpha_total += th.alpha;
        }

        // Numerical stabilisation
        to->setStable(ct = (double)1.0/alpha_total);
        for (tc = 0; tc < to_ntags ; ++tc)
        {
            FBDatum &th = to->getData(tc);
            th.alpha = (th.alpha * ct) + TagBase::tinyScore;
        }

        from = to;
    }
}

void Bigram::FBTagger::beta_pass()
{
    TagChart::reverse_iterator i = chart.rbegin();
    FBSpine *to = (FBSpine *) *i;

    // Initialise scores on last node
    to->getData(0).beta  = 1;

    /* Main body of pass */
    for (i++; i != chart.rend(); i++)
    {
        FBSpine *from = (FBSpine *) *i;

        unsigned int fc, tc; // from and to counters

        unsigned int from_ntags = from->tagCount();
        unsigned int to_ntags   = to->tagCount();

        for (fc = 0; fc < from_ntags ; ++fc)
        {
            Lexicon::LexTag &fromTag = from->getTag(fc);
            FBDatum &fh = from->getData(fc);
            fh.beta = TagBase::tinyScore;

            for (tc = 0; tc < to_ntags ; ++tc)
            {
                Lexicon::LexTag &toTag = to->getTag(tc);
                fh.beta += to->getData(tc).beta * fromTag.transScore(toTag)
                    * toTag.score + TagBase::tinyScore;
            }

            // Numerical stabilisation
            fh.beta = (fh.beta * to->getStable()) + TagBase::tinyScore;
        }

        to = from;
    }
}

void Bigram::FBTagger::gamma_pass()
{
    for (TagChart::iterator i = chart.begin(); i != chart.end(); i++)
    {
        FBSpine *node = (FBSpine *) *i;

        unsigned int c;
        double best = 0;

        // Calculate overall score. You would normally divide by the
        // observation sequence probability here, but the numerical
        // stabilisation means this this will always be 1. We also choose the
        // hypotheses at this stage
        unsigned int ntags = node->tagCount();
        for (c = 0; c < ntags ; ++c)
        {
            FBDatum &h = node->getData(c);
            h.gamma = (h.alpha * h.beta) + TagBase::tinyScore;
            if (h.gamma > best)
            {
                best = h.gamma;
                node->setChosen(c);
            }
        }
    }
}

//---------------------------------------------------------------------------
// Viterbi
void Bigram::ViterbiTagger::tag()
{
    unsigned int chosen;
    unsigned int fc, tc; // from and to counters

    TagChart::iterator i = chart.begin();
    ViterbiSpine *from = (ViterbiSpine *) *i;

    // Initialise scores on first node
    from->getData(0).score = from->getTag(0).score;

    // Main body of pass
    for (i++; i != chart.end(); i++)
    {
        ViterbiSpine *to = (ViterbiSpine *) *i;
        
        unsigned int to_ntags = to->tagCount();
        for (tc = 0; tc < to_ntags ; ++tc)
        {
            Lexicon::LexTag &toTag = to->getTag(tc);
            ViterbiDatum &th = to->getData(tc);
            th.score = 0;
            th.link  = 0;

            unsigned int from_ntags = from->tagCount();
            
            for (fc = 0; fc < from_ntags ; ++fc)
            {
                ViterbiDatum &fh = from->getData(fc);
                double s = fh.score * toTag.score * from->getTag(fc).transScore(toTag)
                    + TagBase::tinyScore;

                if (s > th.score)
                {
                    th.score = s;
                    th.link  = fc;
                }
            }
        }

        from = to;
    }

    // Assign the chosen values
    TagChart::reverse_iterator j = chart.rbegin();
    ViterbiSpine *to = (ViterbiSpine *) *j;
    to->setChosen(chosen = 0);

    for (j++; j != chart.rend(); j++)
    {
        ViterbiSpine *from = (ViterbiSpine *) *j;
        from->setChosen(chosen = to->getData(chosen).link);
        to = from;
    }
}
