#ifndef Bigram_H
#define Bigram_H
// Bigram tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagbase.h"
#include "lexicon.h"
#include "endings.h"

// The topm level Bigram class really serves as a namespace, with the actual
// classes underneath it.
class Bigram
{
    public:
    // Some useful utilities.

    // Write/read the entire collection of data needed for bigram
    // tagging. Read creates the word rules and returns false on error.
    // read assumes binary; write can do binary or testial.
    static bool readData
        (istream &in, TagSet &tagset, Lexicon &lex, Endings &rules);
    static void writeData
        (ostream &f, const TagSet &tagset, const Lexicon &lex,
         const Endings &rules, bool textual);


    //------------------------------------------------------------------------
    // Data structures to represent the chart used in tagging. The basic data
    // structure is TagSpine, which is a node of a list. Each node represents a
    // token and its properties. Classes for specific tagging algorithms are
    // derived from this class. To save some reallocation of memory, we do not
    // necessarily delete this structure when we have finished tagging a sequence,
    // but may reuse it.

    class TagSpine
    {
        public:
        // Create an empty tagspine node.
        TagSpine() : lex(NULL), local(false), tags(NULL) {}

        // Delete a tagspine node.
        virtual ~TagSpine()
        {
            clear_node();
        }

        // Fill a node from a lexical entry, applying endings rules if unknown.
        void fill_node(Lexicon::LexEntry &lex, const Endings &rules);

        // Clear a node, i.e. reset its fields
        void clear_node()
        {
            clearer();
            if (local && tags != NULL)
            {
                delete tags;
            }
            tags = NULL;
        }

        // Get the lexical entry
        const Lexicon::LexEntry *getLex() const { return lex; }

        // Get number of tags on this node
        unsigned int tagCount() const
        {
            return tags->size();
        }

        // Get the ith tag
        Lexicon::LexTag &getTag(unsigned int i) const
        {
            return (*tags)[i];
        }

        // Get chosen tag index
        unsigned int getChosen() const { return chosen; }

        // Set chosen tag index
        void setChosen(unsigned int c) { chosen = c; }

        // Access the token
        TaggedToken &getToken() { return token; }

        // Access the token
        const TaggedToken &getToken() const { return token; }

        protected:
        // The token
        TaggedToken token;

        // The lexical entry for token.
        Lexicon::LexEntry *lex;

        // Index of the chosen tag.
        unsigned int chosen;

        // Pointer to the tags. This can either be the tags belonging to the lexical
        // entry, or a local copy. The latter is used in the case where endings rules
        // modify the tags. The flag says whether it is local or not.
        bool local;
        Lexicon::LexTags *tags;

        // Fill in a node, specific to a tagging algorithm.
        virtual void filler() =0;

        // Clear a node, specific to a tagging algorithm.
        virtual void clearer() {}
    };

    // Class for Forward-Backward representation.
    // alpha, beta and gamma are the forward, backward and product scores.
    struct FBDatum
    {
        FBDatum() : alpha(0), beta(0), gamma(0) {}
        double alpha, beta, gamma;
    };

    // Node class when doing forward-backward tagging.
    class FBSpine : public TagSpine
    {
        public:
        // Create a FB spine node with empty data.
        FBSpine() {}

        // Free a FB spine node.
        ~FBSpine() {}

        // Get the ith data element
        FBDatum &getData(unsigned int i)
        {
            return data[i];
        }

        // Get the ith data element
        const FBDatum &getData(unsigned int i) const
        {
            return data[i];
        }

        // Set the numerical stabilization
        void setStable(double s) { ct = s; }

        // Get the numerical stabilization
        double getStable() const { return ct; }

        private:
        // The data. This array parallels the tags.
        typedef vector<FBDatum> FBData;
        FBData data;

        // Numerical stabilisation parameter
        double ct;

        void filler();
        void clearer();
    };

    // Class for Viterbi representation.
    // It has a score, and the index of the tag on the preceding node.
    struct ViterbiDatum
    {
        ViterbiDatum() : score(0), link(0) {}

        double score;
        unsigned int link;
    };

    // Node class when doing Viterbi tagging.
    class ViterbiSpine : public TagSpine
    {
        public:
        // Create Viterbi spine node
        ViterbiSpine() {}

        // Destroy Viterbi spine node
        ~ViterbiSpine() {}
        
        // Get the ith data element
        ViterbiDatum &getData(unsigned int i)
        {
            return data[i];
        }

        private:
        // The data. This array parallels the tags.
        typedef vector<ViterbiDatum> ViterbiData;
        ViterbiData data;

        void filler();
        void clearer();
    };

    //------------------------------------------------------------------------
    // The main class for taggers themselves.
    // Specific tagging algorithms are derived from this.
    class Tagger
    {
        protected:
        // The overall chart for tagging consists of a list of nodes.
        typedef list<TagSpine *> TagChart;

        public:
        // Construct a tagger.
        Tagger() {}

        // Destroy a tagger.
        virtual ~Tagger() {}

        // Do the tagging
        virtual void tag() =0;

        //--------------------------------------------------------------------
        // Top level tagging function. Calls token_handler, unknown_handler,
        // writer. The initial flag says whether to use initial node in output:
        // usually false if an anchor was retained. The casehold flag is passed in
        // to indicate whether we preserve case; case transformations may be made if
        // a lexicon lookup fails.
        void tagger
            (const TokenReader *reader, const Lexicon &lex, const TagSet &tagset,
             const Endings &rules, istream &in, bool casehold, bool initial);

        protected:
        // Tear down a chart after tagging, retaining an anchor
        void clear_chart();

        // Destroy chart, including spine
        void destroy_chart();

        // Create a new chart from scratch (old one, if any, is lost), seeding
        // it with an anchor.
        void new_chart(const Lexicon::LexEntry &anchor);

        // Do extra processing on tokens in tagging. May modify the token.
        virtual void token_handler(TaggedToken &) {}

        // Extra processing for unknown words.
        virtual void unknown_handler(const Lexicon &, const TaggedToken &) {}

        // Write the chart at the end of each tagging sequence.
        // The initial flag is the one passed to tagger(). If set, the first word
        // should be skipped.
        virtual void writer(bool initial) const =0;

        protected:
        // The tagger chart, i.e. where the nodes are stored.
        TagChart chart;

        // Create a node
        virtual TagSpine *new_node() =0;
    };

    // Tagger using the Forward-Backward algorithm.
    class FBTagger : public Tagger
    {
        public:
        // Create FB tagger.
        FBTagger() {}

        // Destroy FB tagger
        ~FBTagger() {}

        // Execute the tagging.
        void tag();

        private:
        // Create a new node.
        TagSpine *new_node() { return new FBSpine; }

        // Passes of the tagging algorithm.
        void alpha_pass();
        void beta_pass ();
        void gamma_pass();
    };

    // Tagger using the Viterbi algorithm.
    class ViterbiTagger : public Tagger
    {
        public:
        // Create Viterbi tagger.
        ViterbiTagger() {}

        // Destroy Viterbi tagger.
        ~ViterbiTagger() {}
        
        // Execute the tagging.
        void tag();

        private:
        // Create a new node.
        TagSpine *new_node() { return new ViterbiSpine; }
    };
};
#endif // Bigram_H
