// Baum-Welch re-estimation
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagset.h"
#include "bw.h"

// Derived class for BW.

class BW : public Bigram::FBTagger
{
    public:
    // Create BW tagger with a given lexicon and tagset.
    BW(ReestLexicon &l, ReestTagSet &t) : lexicon(l), tagset(t) {}

    // "Write" results: this actually does the re-estimation calculation
    void writer(bool initial) const;

    private:
    // Reference to the lexicon
    ReestLexicon &lexicon;

    // Reference to the tagset and transitions
    ReestTagSet  &tagset;
};


void BaumWelch::makeLexicon(istream &corpus, TokenReader *reader, bool casehold)
{
    // Read the words from the corpus into the following lexicon. If you see a
    // word more than once, just ignore on later readings. The first time it
    // is seen, assign all open class tags, as restricted by the endings rules 
    ReestLexicon newlex;
    TaggedToken token;

    // Read words from the corpus and look them up
    while (true)
    {
        Lexicon::LexEntry *lexeme;

        if (!reader->read(corpus, token))
        {
            break;
        }
        TagBase::normalise(token.getToken(), lexicon.getNumberTest(), true);

        // Look for the word in the existing lexicon
        lexeme = lexicon.find(token.getToken());
        if (lexeme->isUnknown() && !casehold)
        {
            TagBase::lowercase(token.getToken());
            lexeme = lexicon.find(token.getToken());
        }

        if (lexeme->isUnknown())
        {
            // Check new lexicon to see if we already have it
            Lexicon::LexEntry *lex_entry = newlex.find(token.getToken());

            if (lex_entry == NULL)
            {
                // Not found: set up tags and add to the new lexicon
                // Make a local copy of the data from the unknown word entry
                unsigned int ntags = lexeme->tagCount();
                Lexicon::LexTags tags;
                lexeme->copyTags(tags);
                ntags = rules.apply(token, tags);

                if (ntags == 0)
                {
                    tags.clear();
                    lexeme->copyTags(tags);
                    ntags = tags.size();
                }

                // Now we can add it to the new lexicon
                lex_entry = newlex.findOrCreate(token.getToken());
            
                for (unsigned int t = 0 ; t < ntags ; ++t)
                {
                    lex_entry->addTag(tags[t].tag, tags[t].score);
                }
            }
        }
    }

    // Merge new lexicon into main one
    lexicon.merge(newlex, tagset);

    // Zero all the re-estimation fields in lexicon and transitions
    lexicon.commit(true);
    tagset.reset();
}

void BW::writer(bool initial) const
{
    for (TagChart::const_iterator ch = chart.begin(); ch != chart.end(); )
    {
        if (ch == chart.begin() && !initial)
        {
            continue;
        }

        // Cast away constness as we need to alter lexicon etc.
        Bigram::FBSpine &node = (Bigram::FBSpine &)**ch;

        // Check whether there is a next node
        bool next = (++ch == chart.end());

        unsigned int ntags = node.tagCount();
        ReestLexicon::ReestLexEntry *lex
            = (ReestLexicon::ReestLexEntry *)(node.getLex());
        for (unsigned int i = 0 ; i < ntags ; ++i)
        {
            // Lexical re-estimation
            lex->addToScore(i, node.getData(i).gamma);

            // Transition re-estimation to next node
            if (next)
            {
                ReestTagSet::ReestTag *stag
                    = (ReestTagSet::ReestTag *)(node.getTag(i).tag);
                Bigram::FBSpine &n = (Bigram::FBSpine &)**ch;
                double alpha = node.getData(i).alpha;

                unsigned int n_ntags = n.tagCount();
                for (unsigned int j = 0 ; j < n_ntags ; ++j)
                {
                    Lexicon::LexTag nTag = n.getTag(j);
                    TagSet::Tag *toTag = n.getTag(j).tag;
                    double beta = n.getData(j).beta;

                    double reest_score = alpha
                        * node.getTag(i).transScore(nTag)
                        * nTag.score * beta * n.getStable();

                    stag->addTo(toTag, reest_score, beta);
                }
            }
        }
    }
}

void BaumWelch::reestimate
(istream &corpus, TokenReader *reader, bool casehold)
{
    BW bw(lexicon, tagset);
    bw.tagger(reader, lexicon, tagset, rules, corpus, casehold, true);
}

void BaumWelch::consolidate()
{
    lexicon.commit(false);
    tagset.commit();
    tagset.reset();
}
