#ifndef BW_H
#define BW_H
// Baum-Welch re-estimation
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagbase.h"
#include "bigram.h"

class BaumWelch
{
    /* Main Baum Welch re-estimator. Takes a bigram tagger control block which
       includes an existing model. It operates in three phases. In the first
       phase, corpora are passed in and the words accumulated into a new
       lexicon. The tags come from a combination of the tags
       list and the ending rules present in the model. At the end of the first
       phase, the new lexicon is merged with the main lexicon. In the
       second phase, the corpus is re-read, tagged, and the re-estimation
       results accumulated. Finally, in the third phase, the results are
       consolidated back into the original model. This may involve a certain
       amount on (apparently) unnecessary copying things around, but since you
       won't do re-estimation all that often, the penalty in time is not very
       important.
    */

    public:
    // Construct over a given tagset, lexicon and endings rules.
    BaumWelch(ReestTagSet &t, ReestLexicon &l, const Endings &e)
        : tagset(t), lexicon(l), rules(e) {}

    // First phase: make a raw lexicon from the corpus.
    void makeLexicon(istream &corpus, TokenReader *reader, bool casehold);

    // Second phase: apply re-estimation.
    void reestimate(istream &corpus, TokenReader *reader, bool casehold);

    // Final consolidation of results
    void consolidate();

    private:
    // Tagset and transitions
    ReestTagSet &tagset;

    // Lexicon
    ReestLexicon &lexicon;

    // Endings rules for unknown words
    const Endings &rules;
};

#endif // BW_H

