// Baum-Welch trainer
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "bigram.h"
#include "bw.h"
#include "getopt.h"
#include <fstream>
#include <cstdlib>
using namespace std;

const unsigned int DefaultIterations = 10;

int main(int argc, char *argv[])
{
    // Input formats
    enum Format { Penn, Tabbed, Untagged, Unknown, NotGiven };
    string outname;
    
    ofstream outfile;
    
    unsigned int iterations = DefaultIterations;
    const TagBase::NumberTest *number = TagBase::AnyDigit;
    
    bool ok = true;
    bool have_datafile = false;
    bool reduce_lex = false;
    unsigned int reduce_level = 0;

    Format input_format = Untagged;
    bool casehold = false, textual = false;

    ReestTagSet  tagset;
    ReestLexicon lexicon;
    Endings rules;
    BaumWelch bw(tagset, lexicon, rules);

    int c;

    opterr = 0;

    // Verify the arguments
    while ((c = local_getopt(argc, argv, "o:i:lI:m:nNcr:T")) != -1)
    {
        switch (c)
        {
            case 'm':
                if (have_datafile)
                {
                    cerr << "Datafile already read: " << optarg
                        << " ignored" << endl;
                }
                else
                {
                    ifstream d(optarg, ios::in | ios::binary);

                    if (!d)
                    {
                        cerr << "Cannot open datafile " << optarg << endl;
                    }
                    else
                    {
                        have_datafile = Bigram::readData
                            (d, tagset, lexicon, rules);
                    }
                }
                break;
            case 'o':
                if (outname != "")
                {
                    cerr << "Output file specified more than once" << endl;
                    ok = false;
                }
                else
                {
                    outname = optarg;
                }
                break;
            case 'i':
                if (strcmp(optarg, "p") == 0)
                    input_format = Penn;
                else if (strcmp(optarg, "t") == 0)
                    input_format = Tabbed;
                else if (strcmp(optarg, "u") == 0)
                    input_format = Untagged;
                else
                {
                    cerr << "Unknown input format '" << optarg
                        << "': using default" << endl;
                    input_format = Untagged;
                }
                break;
            case 'I':
                iterations = atoi(optarg);
                if (iterations <= 0)
                {
                    iterations = DefaultIterations;
                }
                break;
            case 'n':
                number = TagBase::AllDigit;
                break;
            case 'N':
                number = TagBase::ParseableNumber;
                break;
            case 'l':
                casehold = true;
                break;
            case 'T':
                textual = true;
                break;
            case 'r':
                reduce_lex = true;
                reduce_level = atoi(optarg);
                break;
            case '?':
                cerr << "Bad or unknown option " << argv[optind-1] << endl;
                break;
        }
    }

    if (outname == "")
    {
        cerr << "An output file must be specified" << endl;
        ok = false;
    }
 
    if (input_format == NotGiven)
    {
        cerr << "An input format must be specified" << endl;
        ok = false;
    }
    
    if (ok && optind == argc)
    {
        cerr << "No input files specified" << endl;
        ok = false;
    }
    
    if (ok)
    {
        if (!textual)
        {
            outfile.open(outname.c_str(), ios::out | ios::binary);
        }
        else
        {
            outfile.open(outname.c_str());
        }
        
        if (!outfile)
        {
            cerr << "Cannot open output file '" << outname << "'" << endl;
            ok = false;
        }
    }

    if (!ok)
    {
        cerr << "Error(s) in command line" << endl;
        return -1;
    }
    else
    {
        bool any = false;
        
        TokenReader *reader
            = (input_format == Penn)
            ? TokenReader::reader_penn(tagset)
            : (input_format == Tabbed)
            ? TokenReader::reader_tabbed(tagset)
            : (input_format == Untagged)
            ? TokenReader::reader_untagged()
            : NULL;

        // First pass through the corpora, to pick up vocabulary
        for (int i = optind ; i < argc ; i++)
        {
            ifstream in(argv[i]);
            if (!in)
            {
                cerr << "Cannot open input file '" << argv[i] << "'" << endl;
            }
            else
            {
                any = true;
                bw.makeLexicon(in, reader, casehold);
                in.close();
            }
        }

        if (!any)
        {
            cerr << "No input files were valid" << endl;
            return -1;
        }

        // Second pass to do the reestimation
        while (iterations-- > 0)
        {
            for (int i = optind ; i < argc ; i++)
            {
                ifstream in(argv[i]);
                if (!in)
                {
                    bw.reestimate(in, reader, casehold);
                    in.close();
                }
            }

            // Consolidation for next pass
            bw.consolidate();
            tagset.transNormalise();
            lexicon.normalise();
        }

        // Apply reduction
        if (reduce_level > 0)
        {
            lexicon.reduce(reduce_level);
        }
        
        // Final write
        Bigram::writeData(outfile, tagset, lexicon, rules, textual);
        delete reader;

        return 0;
    }
}
