// Endings and other pseudo-morphology for tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "endings.h"
#include <cctype>
#include <sstream>
#include <limits>
using namespace std;

static unsigned int MaxInt = numeric_limits<unsigned int>::max();

//--------------------------------------------------------------------------
// Get a positive number from a stream, with some error handling
// After reading successfully, it leaves the next character in c
static bool getNumber(istream &ss, unsigned int &result, const string &kind, char &c)
{
    result = MaxInt;
    if (ss.good())
    {
        ss >> result;
    }
    if (ss.good())
    {
        ss >> c;
        return true;
    }
    else
    {
        cerr << "Bad " << kind << " number: " << result << endl;
        return false;
    }
}

bool Endings::readFile(istream &in, const TagSet &tagset, char fp_sep)
{
    rules.clear();
    bool error = false;
    string buffer;

    while (TagBase::readBuffer(buffer, in))
    {
        // Convert the input to a stream
        istringstream ss(buffer);

        rules.push_back(WordRule());
        WordRule &rule = rules.back();

        // Skip any leading white space, and see if we have a number after
        char c;
        ss >> c;
        if (isdigit(c))
        {
            // Get rule number
            ss.putback(c);
            if (!getNumber(ss, rule.number, "rule", c))
            {
                error = true;
                continue;
            }
            ss >> c;
        }

        // Get a continuation number if present
        if (c == ':')
        {
            if (!getNumber(ss, rule.cont, "continuation", c))
            {
                error = true;
                continue;
            }
            else
            {
                ss >> c;
            }
        }

        // Now we are at the pattern. First check for negation
        if (c == '~')
        {
            rule.negated_pattern = true;
        }
        else
        {
            ss.putback(c);
        }

        if (!ss.good())
        {
            cerr << "Missing pattern in " << buffer << endl;
            error = true;
        }
        else
        {
            bool special = true;

            string pattern;
            ss >> pattern;
            string::size_type pstart = 0;
            string::size_type pend   = pattern.length();

            if (pattern == "I")
            {
                rule.type = InitialCap;
            }
            else if (pattern == "A")
            {
                rule.type = AllCap;
            }
            else if (pattern == "C")
            {
                rule.type = AnyCap;
            }
            else if (pattern[0] == '-')
            {
                pstart += 1;
                rule.type = Suffix;
                special = false;
            }
            else if (pattern[pend-1] == '-')
            {
                pend -= 1;
                rule.type = Prefix;
                special = false;
            }
            else
            {
                cerr << "Bad pattern string in " << buffer << endl;
                error = true;
            }

            if (special)
            {
                rule.match_pattern = rule.nomatch_pattern = "";
            }
            else if (!error)
            {
                if (isalpha(pattern[pstart]) && !islower(pattern[pstart]))
                {
                    cerr << "Rule has wrong format in " << buffer << endl;
                    error = true;
                }
                else
                {
                    // Divide the pattern at the bar character, if it has one
                    string::size_type bar = pattern.find('|', pstart);
                    string left_part, right_part;
                    if (bar != pattern.npos)
                    {
                        left_part  = pattern.substr(pstart, bar-pstart);
                        right_part = pattern.substr(bar+1, pend-bar-1);
                    }
                    else
                    {
                        left_part = pattern.substr(pstart, pend-pstart);
                    }

                    if (left_part == "" && right_part == "")
                    {
                        cerr << "Empty pattern in " << buffer << endl;
                        error = true;
                    }

                    if (!error)
                    {
                        if (rule.type == Prefix)
                        {
                            rule.match_pattern   = left_part;
                            rule.nomatch_pattern = right_part;
                        }
                        else
                        {
                            if (right_part != "")
                            {
                                rule.nomatch_pattern = left_part;
                                rule.match_pattern   = right_part;
                            }
                            else
                            {
                                rule.nomatch_pattern = "";
                                rule.match_pattern   = left_part;
                            }
                        }
                    }
                }
            }
        }

        // Scan for the tags
        ss >> c;
        if (c == '~')
        {
            rule.negated_tags = true;
        }
        else
        {
            ss.putback(c);
        }

        rule.tags.clear();
        while (ss.good())
        {
            string tag_name;
            ss >> tag_name;
            if (tag_name == "" )
            {
                break;
            }
            TagSet::Tag *tag;
            double s = 1;

            // Check for score separator
            string::size_type sep = tag_name.find(fp_sep);
            if (sep == tag_name.npos)
            {
                tag = tagset.find(tag_name);
            }
            else
            {
                tag = tagset.find(tag_name.substr(0, sep-1));
                s   = atof(tag_name.substr(sep+1).c_str());
            }

            if (tag == NULL)
            {
                cerr << "Unknown tag in " << tag_name << endl;
                error = true;
            }
            else if (s <= 0)
            {
                cerr << "Bad score " << s << " in " << buffer << endl;
                error = true;
            }
            else
            {
                rule.tags.push_back(TagSet::TagScore(tag, s));
            }
        }
    }
    
    if ( error )
    {
        cerr << "Error(s) in rules file" << endl;
        clear();
        return false;
    }
    else
    {
        return true;
    }
}

//--------------------------------------------------------------------------
void Endings::write(ostream &out) const
{
    unsigned int end = MaxInt;

    // Write each rule in turn, with the number being max at the end
    for (WordRules::const_iterator r = rules.begin(); r != rules.end(); ++r)
    {
        const WordRule &rule = *r;
        rule.write(out);
    }
    out.write((char *)&end, sizeof(end));
}

void Endings::WordRule::write(ostream &out) const
{
    out.write((char *)&number, sizeof(number));
    out.write((char *)&cont,   sizeof(cont));
    out.write((char *)&negated_pattern, sizeof(negated_pattern));
    out.write((char *)&negated_tags, sizeof(negated_tags));
    out.write((char *)&type,  sizeof(type));
    out << match_pattern << '\n';
    out << nomatch_pattern << '\n';

    unsigned int ntags = tags.size();
    out.write((char *)&ntags, sizeof(ntags));
    
    for (Tags::const_iterator i = tags.begin(); i != tags.end(); ++i)
    {
        const TagSet::TagScore &tag = *i;
        out << tag.tag->getName() << '\n';
        out.write((char *)&(tag.score), sizeof(tag.score));
    }
}

bool Endings::read(const TagSet &tagset, istream &in)
{
    rules.clear();
    bool error = false;

    while (in.good())
    {
        unsigned int number;

        in.read((char *)&number, sizeof(number));
        if (!in.good())
        {
            error = true;
            break;
        }
        else if (number == MaxInt)
        {
            break;
        }

        rules.push_back(WordRule());
        WordRule &rule = rules.back();
        rule.number = number;
        if (!rule.read(in, tagset))
        {
            error = true;
            break;
        }
    }

    if (error)
    {
        cerr << "Error(s) in rule data" << endl;
        clear();
        return false;
    }
    else
    {
        return true;
    }
}

bool Endings::WordRule::read(istream &in, const TagSet &tagset)
{
    in.read((char *)&cont, sizeof(cont));
    in.read((char *)&negated_pattern, sizeof(negated_pattern));
    in.read((char *)&negated_tags, sizeof(negated_tags));
    in.read((char *)&type, sizeof(type));
    if (!in.good())
    {
        return false;
    }

    if (!TagBase::readBuffer(match_pattern, in)
        || !TagBase::readBuffer(nomatch_pattern, in))
    {
        return false;
    }

    unsigned int ntags;
    in.read((char *)&ntags, sizeof(ntags));
    if (!in.good())
    {
        return false;
    }

    tags.reserve(ntags);

    for (unsigned int i = 0 ; i < ntags ; ++i)
    {
        string tag_name;
        double score;
        
        if (!TagBase::readBuffer(tag_name, in))
        {
            return false;
        }
        in.read((char *)&score, sizeof(score));
        if (!in.good())
        {
            return false;
        }

        tags.push_back(TagSet::TagScore(tagset.find(tag_name), score));
    }
    return true;
}

void Endings::clear()
{
    rules.clear();
}

unsigned int Endings::apply(const TaggedToken &token_, Lexicon::LexTags &ltags) const
{
    unsigned int from, to;
    bool any_change = false;
    const string &token = token_.getToken();

    if (token == "")
    {
        return ltags.size();
    }

    unsigned int nltags = ltags.size();
    for (WordRules::const_iterator r = rules.begin(); r != rules.end(); )
    {
        const WordRule &rule = *r;
        bool match = false;
        string::size_type mlen = rule.match_pattern.length();
        string::size_type nlen = rule.nomatch_pattern.length();
        string::size_type tlen = token.length();

        if (rule.type == Prefix)
        {
            string::size_type t = 0;
            if (mlen != 0)
            {
                match = TagBase::casecmp(token, rule.match_pattern, mlen);
                t += mlen;
            }

            if (match)
            {
                if (nlen != 0)
                {
                    match = !TagBase::ncasecmp(token, rule.nomatch_pattern, nlen, t);
                }
            }
        }
        else if (rule.type == Suffix)
        {
            string::size_type end = tlen;
            if (mlen != 0 && end > mlen)
            {
                end -= mlen;
                match = TagBase::casecmp(token, rule.match_pattern, end);
            }

            if (match)
            {
                if (nlen != 0)
                {
                    match = !TagBase::ncasecmp
                        (token, rule.nomatch_pattern, nlen, end-nlen);
                }
            }
        }
        else if (rule.type == InitialCap)
        {
            for (unsigned int l = 0 ; l < tlen ; ++l)
            {
                if (isalpha(token[l]))
                {
                    match = isupper(token[l]);
                    break;
                }
            }
        }
        else if (rule.type == AllCap)
        {
            unsigned int l;
            bool any_alpha = false;
            for (l = 0 ; l < tlen ; ++l)
            {
                if (isalpha(token[l]))
                {
                    any_alpha = true;
                    if (!isupper(token[l]))
                    {
                        break;
                    }
                }
            }
            if (any_alpha && l == tlen)
            {
                match = true;
            }
        }
        else if (rule.type == AnyCap)
        {
            for (unsigned int l = 0 ; l < tlen ; ++l)
            {
                if (isalpha(token[l]) && isupper(token[l]))
                {
                    match = true;
                    break;
                }
            }
        }

        if ((rule.negated_pattern && !match) ||
            (!rule.negated_pattern &&  match))
        {
            // Have a successful match, so apply the effects of the rule
            unsigned int  nrtags = rule.tags.size();

            // Scan the tag list and unhook any we do not want
            for (unsigned int l = 0 ; l < nltags ; ++l)
            {
                Lexicon::LexTag &ltag = ltags[l];
                unsigned int r;

                // Check in case tag is already removed
                if (ltag.tag == NULL)
                {
                    continue;
                }

                // Check against each tag in the rule's list
                for (r = 0 ; r < nrtags ; ++r)
                {
                    if (ltag.tag == rule.tags[r].tag)
                    {
                        break;
                    }

                    // Remove if not in list and not negated, or vice-versa
                    if ((r == nrtags && !rule.negated_tags) ||
                        (r != nrtags &&  rule.negated_tags))
                    {
                        ltag.tag = NULL;
                        any_change = true;
                    }
                    else
                    {
                        // Not removed. If rule has a score, bias with it
                        if (r != nrtags && rule.tags[r].score >= 0)
                        {
                            ltag.score *= rule.tags[r].score;
                        }
                    }
                }
            }
        }
        
        // If there is continuation, skip to it; otherwise, get out
        if (rule.cont > 0)
        {
            unsigned int cont = rule.cont;
            for (++r; r != rules.end(); ++r)
            {
                if ((*r).cont == cont)
                {
                    break;
                }
            }
        }
        else
        {
            ++r;
        }
    }

    // Compact the tags
    if (any_change)
    {
        for (from = to = 0 ; from < nltags ; ++from)
        {
            if (ltags[from].tag != NULL)
            {
                if (from != to)
                {
                    ltags[to++] = ltags[from];
                }
                else
                {
                    to += 1;
                }
            }
        }
    }
    else
    {
        to = nltags;
    }

    return to;
}
