#ifndef Endings_H
#define Endings_H
// Endings and other pseudo-morphology for tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "lexicon.h"
#include "token.h"
#include <vector>
#include <list>
using namespace std;

/* Rules have the form:
  [<num>][<:num>][~]<pattern> [~]<tags>
  where <tags> is a list of tags separated by spaces. The rule means if the
  word meets (~ => fails to meet) the pattern, then restrict the tags to the
  given ones (~ => eliminate the listed tags). If the rules fires and a :<num>
  is given, then jump to the rule with label <num>. If the rules fires and
  there is no such continuation, no further rules are tested. Error checking
  is fairly weedy, so if the continuation does not exist, then nothing happens.
  It is up to the rule set author to make sure they appear in a sensible
  order. In particular, and all caps rule ought to preced an initial cap or
  any cap rule.

  Patterns can be:
     suffixes -x|y matching a word ending in y but not xy.
     prefixes y|x- matching a word starting in y but not yx.
     If x is empty, may omit |. y may be empty if x is not. Comparisons are
     not case sensitive.
     Special patterns: I for initial capital, A for all capitals, C for any
     capital.

  Each tag may be followed by a floating point number to give an initial bias
  to the probability. The number is joined on to the tag by a separator given
  in the read call.

  This is a very crude mechanism, but can be put to good use.
*/

class Endings
{
    protected:
    // An array of tags with scores
    typedef vector<TagSet::TagScore> Tags;

    // Enumeration of rule types. This is not a very object-oriented way of doing things,
    // but it will do.
    enum RuleType {Prefix, Suffix, InitialCap, AllCap, AnyCap};

    // The word rule structure
    struct WordRule
    {
        // Create an empty word rule
        WordRule() : number(0), cont(0), negated_pattern(false), negated_tags(false)
        {}

        // Destroy word rule.
        ~WordRule() {}

        // Read in binary format. Return false on a read error.
        bool read(istream &is, const TagSet &tagset);

        // Write in binary format.
        void write(ostream &os) const;

        // Id number for this rule, or 0
        unsigned int number;

        // Id number for next rule, or 0
        unsigned int cont;

        // Flags to indicate pattern and/or tags are negated
        bool negated_pattern, negated_tags;

        // Type of the rule
        RuleType type;

        // Part of pattern that must match
        string   match_pattern;      

        // Part of pattern that must not match
        string   nomatch_pattern;

        // Array of tags and scores
        Tags     tags;
    };

    public:
    // Create empty endings rules.
    Endings() {}

    // Destroy endings rules
    ~Endings() { clear(); }

    // Report if there are any rules
    bool haveRules() const { return rules.size() > 0; }

    // Read a file of endings, textually.
    // fp_sep is a separator between a tag and a score.
    bool readFile(istream &in, const TagSet &tagset, char fp_sep);

    // Write a file of endings, binary format.
    void write(ostream &out) const;

    // Read endings data from a file, binary format.
    // Return false on a read error.
    bool read(const TagSet &tagset, istream &in);

    // Free the rules.
    void clear();

    // Take an array of lexical tags and a surface form and apply the
    // rules. The number of tags left is returned, and the array is
    // compacted. Note that this could result in all the tags being
    // eliminated, in which case the caller must take appropriate action.
    // NB! ltags is destructively modified, so make sure it is not a pointer
    // into the lexicon. 
    unsigned int apply(const TaggedToken &token, Lexicon::LexTags &ltags) const;

    private:
    // List of word rules
    typedef list<WordRule> WordRules;
    WordRules rules;
};
#endif // Endings_H
