// Lexicon class for tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "lexicon.h"

static string read_err = "Lexicon read error";

//-----------------------------------------------------------------------------------
bool Lexicon::LexEntry::readTags(unsigned int ntags, const TagSet &tagset, istream &in)
{
    tags.clear();

    for (unsigned int j = 0 ; j < ntags ; ++j)
    {
        string tagname;
        if (!TagBase::readBuffer(tagname, in))
        {
            cerr << read_err << endl;
            return false;
        }

        double score;
        in.read((char *)&score, sizeof(score));
        if (!in.good())
        {
            cerr << read_err << endl;
            return false;
        }

        addTag(tagset.find(tagname), score);
    }
    return true;
}

void Lexicon::LexEntry::writeTags(ostream &out, bool textual) const
{
   unsigned int ntags = tags.size();

    if (textual)
    {
        for (LexTags::const_iterator j = tags.begin(); j != tags.end(); ++j)
        {
            const LexTag &tag = *j;
            if (j != tags.begin())
            {
                out << ' ';
            }
            out << tag.tag->getName() << ' ' << tag.score;
        }
    }
    else
    {
        out.write((char *)&ntags, sizeof(ntags));
        for (LexTags::const_iterator j = tags.begin(); j != tags.end(); ++j)
        {
            const LexTag &tag = *j;
            out << tag.tag->getName() << '\n';
            
            double score = tag.score;
            out.write((char *)&score, sizeof(score));
        }
    }
}

void Lexicon::LexEntry::normalise()
{
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j)
    {
        LexTag &tag = *j;

        double total = tag.tag->getTotal();
        if (total <= TagBase::tinyScore)
        {
            tag.score = TagBase::tinyScore; // Defensive case
        }
        else
        {
            tag.score /= total;
        }
    }
}

void Lexicon::LexEntry::denormalise()
{
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j)
    {
        LexTag &tag = *j;
        tag.score *= tag.tag->getTotal();
    }
}

// This could probably be improved by keeping the tags in sorted order.
// However, most words have few tags, so it really doesn't make much difference.
void Lexicon::LexEntry::addTag(TagSet::Tag *tag, double score)
{
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j)
    {
        LexTag &ltag = *j;

        if (ltag.tag == tag)
        {
            ltag.score += score;
            return;
        }
    }

    newTag(tag, score);
}


void Lexicon::LexEntry::removeTagScore()
{
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j)
    {
        LexTag &tag = *j;
        tag.tag->addToTotal(-tag.score);
    }
}

double Lexicon::LexEntry::getTotalScore(bool isNormalised)
{
    double total = 0;
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j)
    {
        LexTag &tag = *j;
        double score = tag.score;
        if (isNormalised)
        {
            score *= tag.tag->getTotal();
        }
        total += score;
    }

    return total;
}

void Lexicon::LexEntry::addAllTags(const LexEntry *source, TagSet &tagset)
{
    for (LexTags::const_iterator j = source->tags.begin(); j != source->tags.end(); ++j)
    {
        const LexTag &tag = *j;
        addTag(tagset.find(tag.tag->getName()), tag.score);
    }
}


//-----------------------------------------------------------------------------------

// Read a lexicon from a file (binary); not cleared first
/* Format:
    - code for number test
    - lexical entries:
    <nentries>
    [<word><ntags>[<tag name><score>]+]+
    Unknown and anchor words are written just like any other entry
*/
bool Lexicon::read(const TagSet &tagset, istream &in)
{
    unsigned int numberCode;
    in.read((char *)&numberCode, sizeof(numberCode));
    if ( !in.good() )
    {
        cerr << read_err << endl;
        return false;
    }
    number = TagBase::getNumberTest(numberCode);

    // Read entries
    unsigned int nlex;
    in.read((char *)&nlex, sizeof(nlex));
    if ( !in.good() )
    {
        cerr << read_err << endl;
        return false;
    }

    for (unsigned int i = 0 ; i < nlex ; ++i)
    {
        string word;
        if (!TagBase::readBuffer(word, in))
        {
            cerr << read_err << endl;
            return false;
        }

        LexEntry *lexe = makeEntry(word);
        lex[word] = lexe;
        
        unsigned int ntags;
        in.read((char *)&ntags, sizeof(ntags));
        if (!in.good())
        {
            cerr << read_err << endl;
            return false;
        }

        if (!lexe->readTags(ntags, tagset, in))
        {
            return false;
        }
    }

    // Find the anchor and unknown entries
    anchor  = find(TagBase::anchorWord);
    unknown = find(TagBase::unknownWord);
    return true;
}

// Binary format is described above
// Textual format is mainly useful for checking
void Lexicon::write(ostream &out, bool textual) const
{
    if (textual)
    {
        // Output number test as a code.
        unsigned int numberCode = number->code();
        out << "<NUMBERTEST " << numberCode << ">" << endl;

        out << "<LEX>" << endl;
        for (LexMap::const_iterator i = lex.begin(); i != lex.end(); ++i)
        {
            const string &word   = (*i).first;
            const LexEntry *lexe = (*i).second;
            out << word << '\t';

            lexe->writeTags(out, true);
            out << endl;
        }

        out << "</LEX>" << endl;
    }
    else
    {
        unsigned int numberCode = number->code();
        out.write((char *)&numberCode, sizeof(numberCode));

        unsigned int nlex = lex.size();
        out.write((char *)&nlex, sizeof(nlex));
        for (LexMap::const_iterator i = lex.begin(); i != lex.end(); ++i)
        {
            const string &word = (*i).first;
            out << word << '\n';
            
            const LexEntry *entry = (*i).second;
            entry->writeTags(out, false);
        }
    }
}

Lexicon::LexEntry *Lexicon::findOrCreate(const string &word)
{
    LexMap::iterator at = lex.find(word);
    if (at != lex.end())
    {
        return (*at).second;
    }
    else
    {
        LexEntry *entry = lex[word] = makeEntry(word);
        return entry;
    }
}

Lexicon::LexEntry *Lexicon::find(const string &word) const
{
    LexMap::const_iterator at = lex.find(word);
    if (at != lex.end())
    {
        return (*at).second;
    }
    else
    {
        return unknown;
    }
}

void Lexicon::setUnknown(TagSet &tagset, double score)
{
    unknown = findOrCreate(TagBase::unknownWord);

    set<TagSet::Tag *> openTags;
    tagset.openTags(openTags);

    for (set<TagSet::Tag *>::iterator i = openTags.begin(); i != openTags.end(); ++i)
    {
        unknown->addTag(*i, score);
        (*i)->addToTotal(score);
    }
}

void Lexicon::removeUnknown()
{
    if (unknown != NULL)
    {
        unknown->removeTagScore();
    }
}

void Lexicon::adjustAnchor(TagSet &tagset)
{
    // Adjust scores on anchor; creates it if needed.
    if (anchor == NULL)
    {
        anchor = findOrCreate(TagBase::anchorWord);
    }
    anchor->addTag(tagset.getAnchor(), tagset.getAnchor()->getTotal());
}

void Lexicon::normalise()
{
    for (LexMap::iterator i = lex.begin(); i != lex.end(); ++i)
    {
        LexEntry *lexe = (*i).second;
        lexe->normalise();
    }
}

void Lexicon::denormalise()
{
    for (LexMap::iterator i = lex.begin(); i != lex.end(); ++i)
    {
        LexEntry *lexe = (*i).second;
        lexe->denormalise();
    }
}

void Lexicon::clear()
{
    for (LexMap::iterator i = lex.begin(); i != lex.end(); ++i)
    {
        LexEntry *lexe = (*i).second;
        delete lexe;
    }

    lex.clear();
    unknown = anchor = NULL;
}

void Lexicon::reduce(double threshold, bool isNormalised)
{
    // Check for a good threshold and silently fail if not
    if (threshold <= 0)
    {
        return;
    }

    // Now the main scanning loop.
    // We will build a list of all entries which are to be deleted.
    for (LexMap::iterator i = lex.begin(); i != lex.end(); )
    {
        LexEntry *lexe = (*i).second;
        
        if (!lexe->isUnknown() && !lexe->isAnchor())
        {
            double total = lexe->getTotalScore(isNormalised);
            if (total <= threshold)
            {
                // Kill it and get new iterator
                lex.erase(i++);
                continue;
            }
        }

        // Didn't delete
        ++i;
    }
}

void Lexicon::merge(const Lexicon &source, TagSet &tagset)
{
    // Do nothing with unknown and anchor.
    // Take the number test only if we do not already have one
    if (number == NULL)
    {
        number = source.number;
    }

    // Grab each entry.
        for (LexMap::const_iterator l = source.lex.begin(); l != source.lex.end(); ++l)
    {
        const string &word  = (*l).first;
        const LexEntry *src = (*l).second;

        // Find/create destination entry
        LexEntry *dest = findOrCreate(word);
        dest->addAllTags(src, tagset);
    }
}

//----------------------------------------------------------------------------------
void ReestLexicon::ReestLexEntry::commit(bool clear_only)
{
    unsigned int i = 0;
    for (LexTags::iterator j = tags.begin(); j != tags.end(); ++j, ++i)
    {
        LexTag &tag = *j;

        if (!clear_only)
        {
            tag.score = rscores[i] + TagBase::tinyScore;
        }
        rscores[i] = 0.0;
    }
}

void ReestLexicon::commit(bool clear_only)
{
    for (LexMap::iterator i = lex.begin(); i != lex.end(); ++i)
    {
        LexEntry *lexe = (*i).second;
        ((ReestLexEntry *)lexe)->commit(clear_only);
    }
}

