#ifndef Lexicon_H
#define Lexicon_H
// Lexicon class for tagger
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagbase.h"
#include "tagset.h"
#include <map>
using namespace std;

// Main lexicon class.
// A second class follows for the case where re-estimation is being used.
class Lexicon
{
    public:
    // A lexical tag is a tag pointer and a score.
    class LexTag
    {
        public:
        // Construct with a given tag and score
        LexTag(TagSet::Tag *tag_ = TagSet::NoTag, double score_ = TagBase::tinyScore)
            : tag(tag_), score(score_) {}

        // Destroy the tag.
        ~LexTag() {}

        // Get a transition score between two LexTags.
        double transScore(const LexTag &to)
        {
            return tag->tagToTag(to.tag);
        }

        // Get a transition score between two LexTags.
        double transScore(const LexTag *to)
        {
            return tag->tagToTag(to->tag);
        }

        // The tag and score.
        // As this class is really not meant as much more than a pair, we keep
        // these public.
        TagSet::Tag *tag;
        double       score;
    };

    typedef vector<LexTag> LexTags;
    
    // The basic lexical entry consists of a word and its tags.
    // For re-estimation, we will introduce an additional type.

    // Basic class for lexical entry.
    class LexEntry
    {
        public:
        // Create lexical entry with a given word
        LexEntry(const string &word_) : word(word_) {}

        // Destroy the entry
        virtual ~LexEntry() {}

        // Read the tags.
        // There are ntags to read, The tagset is used to find them, and they are
        // read from the binary stream in. Returns true on a successful read.
        bool readTags(unsigned int ntags, const TagSet &tagset, istream &in);

        // Write the tags in either text or binary
        void writeTags(ostream &out, bool textual) const;

        // Normalize tag scores.
        void normalise();

        // Denormalize tag scores.
        void denormalise();

        // Get the word
        const string &getWord() const
        {
            return word;
        }

        // Get number of tags
        // Get number of tags
        unsigned int tagCount() const
        {
            return tags.size();
        }
        
        // Get the ith tag.
        const LexTag &getTag(unsigned int i) const
        {
            return tags[i];
        }

        // Make a copy of the tags.
        void copyTags(LexTags &dest)
        {
            dest.insert(dest.begin(), tags.begin(), tags.end());
        }

        // Make the whole set of tags visible. Use with extreme caution.
        // This is meant mainly for use by the Tagger class.
        LexTags *getTags()
        {
            return &tags;
        }
        
        // Add a tag, or add to the score of an existing tag
        void addTag(TagSet::Tag *tag, double score);
        
        // Add a tag, or add to the score of an existing tag
        void addTag(const TagSet::TagScore &tagScore)
        {
            addTag(tagScore.tag, tagScore.score);
        }
        
        // Return true if this is the unknown word entry.
        bool isUnknown() const
        {
            return (word == TagBase::unknownWord);
        }

        // Return true if this is the anchor word entry.
        bool isAnchor() const
        {
            return (word == TagBase::anchorWord);
        }

        // Remove the tags for this word, keeping tag score totals straight.
        // Mainly meant for the unknown word entry
        void removeTagScore();

        // Get total score, optionally denormalising. Used for reduce().
        double getTotalScore(bool isNormalised);

        // Add all the tags found on another entry.
        // Translate them from one tagset to another as we do so.
        void addAllTags(const LexEntry *source, TagSet &tagset);

        protected:
        // The word.
        string word;

        // Add entry to the tags array. Virtual, so that reest can allocate space.
        virtual void newTag(TagSet::Tag *tag, double score)
        {
            tags.push_back(LexTag(tag, score));
        }
        
        // The tags
        LexTags tags;
    };

    protected:
    // Main part of lexicon representation.
    // Map from lookup keys to pointers to lexical entries.
    typedef map<string, LexEntry *> LexMap;
    LexMap lex;

    // Special entries for the unknown and anchor words
    LexEntry *unknown, *anchor;

    // The number test used for the lexicon.
    const TagBase::NumberTest *number;

    public:
    // Create an empty lexicon
    Lexicon() : unknown(NULL), anchor(NULL), number(NULL) {}

    // Destructor.
    virtual ~Lexicon()
    {
        clear();
    }

    // Read a lexicon from a file (binary); should be cleared first.
    // Tagset is used to resolve tags.
    // Returns false on a read error.
    bool read(const TagSet &tagset, istream &in);

    // Write a lexicon to a file (binary/text)
    void write(ostream &out, bool textual) const;

    // Find a word in a lexicon, creating it if not found, via makeLexEntry().
    LexEntry *findOrCreate(const string &word);

    // Find a word in a lexicon.
    // Yields unknown word if not present and not creating.
    LexEntry *find(const string &word) const;

    // Create an unknown word entry from all open class tags, using given
    // score. Any old one is lost. Do this before normalising.
    void setUnknown(TagSet &tagset, double score);

    // Remove changes in total brought about by the presence of unknown word.
    // Must be un-normalised to do this correctly.
    void removeUnknown();

    // Straightens out the score of the anchor word, so that it is the same as
    // that of the anchor tag.
    void adjustAnchor(TagSet &tagset);

    // Get the anchor entry
    LexEntry *getAnchor() const
    {
        return anchor;
    }

    // Normalise scores in lexicon.
    void normalise();

    // Denormalise scores in lexicon (requires totals to be correct).
    void denormalise();

    // Erase a lexicon, freeing all its memory.
    void clear();

    // Reduce the lexicon according to the threshold.
    // Entries with a total below the threshold are deleted and any remaining
    // pointers to them become invalid.
    // Assumes normalized, can specify not.
    void reduce(double threshold, bool isNormalised = true);

    // Merge another lexicon into this one. Both must be denormalized.
    // Tags are translated to the destination tagset.
    void merge(const Lexicon &source, TagSet &dest_tagset);

    // Get the number test
    const TagBase::NumberTest *getNumberTest() const
    {
        return number;
    }
    
    // Set the number test
    void setNumberTest(const TagBase::NumberTest *n)
    {
        number = n;
    }

    protected:
    // Make a new lexical entry
    virtual LexEntry *makeEntry(const string &word)
    {
        return new LexEntry(word);
    }
};

// Lexicon for use with re-estimation
class ReestLexicon : public Lexicon
{
    public:
    // A re-estimation lexical entry is an ordinary lexical entry together
    // with a vector of re-estimation scores
    class ReestLexEntry : public Lexicon::LexEntry
    {
        typedef vector<double> ReestScores;

        public:
        // Create a lexical entry, with a given word
        ReestLexEntry(const string &word_) : LexEntry(word_) {}

        // Destroy the entry
        virtual ~ReestLexEntry() {}

        // Commit the reestimation values and clear reestimation space. Do before
        // normalisation. Can also be used just for clearing: the function does
        // both because it is more efficient to combine them when both need to be
        // done. 
        void commit(bool clear_only);

        // Add to the score on the ith tag
        void addToScore(unsigned int i, double increment)
        {
            rscores[i] += increment;
        }
        
        private:
        // Re-estimation scores
        ReestScores rscores;

        // Add entry to the tags array.
        void newTag(TagSet::Tag *tag, double score)
        {
            rscores.push_back(0.0);
            tags.push_back(LexTag(tag, score));
        }
    };

    // Create a re-estimation lexicon
    ReestLexicon() {}

    // Destroy a re-estimation lexicon
    ~ReestLexicon() {}

    // Commit the reestimation values and clear reestimation space. Do before
    // normalisation. Can also be used just for clearing: the function does
    // both because it is more efficient to combine them when both need to be
    // done. 
    void commit(bool clear_only);

    protected:
    // Make a new lexical entry
    LexEntry *makeEntry(const string &word)
    {
        return new ReestLexEntry(word);
    }

};

#endif // Lexicon_H
