// Program to read models and merge them
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "train.h"
#include "bigram.h"
#include "getopt.h"
#include <iostream>
#include <fstream>
using namespace std;

// Merge from srcLex into destLex; from srcTagset into destTagset.
// The source tagset and lexicon will be denormalized as we do this, and hence are
// not const.
// Old rules are discarded so do not appear here.
static bool mergeData(TagSet &destTagset, Lexicon &destLex,
                       TagSet &srcTagset,  Lexicon &srcLex)
{
    srcTagset.transDenormalise();
    srcLex.denormalise();

    // When we come to write the results, we will create a new unknown word,
    // so here we remove its influence from the tagset
    srcLex.removeUnknown();

    // Combine the tagsets
    if (!destTagset.mergeIn(srcTagset))
    {
        return false;
    }
    
    // Combine the lexicons
    destLex.merge(srcLex, destTagset);
    
    return true;
}


int main(int argc, char *argv[])
{
    string outname, endname;

    ofstream outfile;
    ifstream endfile;

    Trainer::Correction *corr = Trainer::makeCorrectionNone();

    bool ok = true;
    bool textual = false;
    bool no_norm = false;
    bool reduce_lex = false;
    int reduce_level = 0;

    int c;
    opterr = 0;

    // Get the arguments
    while ((c = local_getopt(argc, argv, "e:o:r:gGTN")) != -1)
    {
        switch (c)
        {
            case 'e':
                if (endname != "")
                {
                    cerr << "Endings already specified: " << optarg
                        << " ignored" << endl;
                }
                else
                {
                    endname = optarg;
                    endfile.open(endname.c_str());
                    if (!endfile.good())
                    {
                        cerr << "Unable to open endings file " << endname
                            << endl;
                        ok = false;
                    }
                }
                break;
            case 'g':
                corr = Trainer::makeCorrectionSGT();
                break;
            case 'G':
                corr = Trainer::makeCorrectionIncrement();
                break;
            case 'o':
                if (outname != "")
                {
                    cerr << "Output already specified: " << optarg
                        << " ignored" << endl;
                }
                else
                {
                    outname = optarg;
                    if (!textual)
                    {
                        outfile.open(outname.c_str(), ios::out | ios::binary);
                    }
                    else
                    {
                        outfile.open(outname.c_str());
                    }
                    
                    if (!outfile.good())
                    {
                        cerr << "Unable to open output file " << outname
                            << endl;
                        ok = false;
                    }
                }
                break;
            case 'T':
                textual = true;
                break;
            case 'N':
                no_norm = true;
                break;
            default:
            case 'r':
                reduce_lex = true;
                reduce_level = atoi(optarg);
                break;
            case '?':
                cerr << "Bad or unknown option " << argv[optind-1] << endl;
                break;
        }
    }

    if (!outfile.good())
    {
        cerr << "A valid output file must be specified" << endl;
        return -1;
    }
    
    if (!ok)
    {
        cerr << "Error(s) in command line" << endl;
        return -1;
    }
 
    if (optind == argc)
    {
        cerr << "No input files specified" << endl;
        return -1;
    }

    TagSet master_tagset;
    Trainer train(master_tagset);

    Lexicon *master_lex = new Lexicon;
    train.setLexicon(master_lex);
    
    for (int i = optind ; i < argc ; ++i)
    {
        ifstream infile(argv[i], ios::in | ios::binary);

        if (!infile.good())
        {
            cerr << "Unable to open model " << argv[i] << endl;
            ok = false;
        }
        else
        {
            TagSet  tagset;
            Lexicon lexicon;
            Endings rules;
            cerr << "Reading " << argv[i] << endl;

            if (!Bigram::readData(infile, tagset, lexicon, rules))
            {
                cerr << "Errors reading " << argv[i] << endl;
                ok = false;
            }
            else
            {
                mergeData(master_tagset, *master_lex, tagset, lexicon);
            }
            infile.close();
        }
    }

    // Create endings, if requested
    Endings out_rules;
    if (endname != "")
    {
        char tagscore_sep = '/';
        if (!out_rules.readFile(endfile, master_tagset, tagscore_sep))
        {
            cerr << "Error in endings file" << endl;
            ok = false;
        }
        endfile.close();
    }

    // Use standard training wrap-up
    if (ok)
    {
        cout << "Writing..." << endl;
        train.end(corr, out_rules, outfile, reduce_level, textual, no_norm);
        outfile.close();
    }

    return (ok) ? 0 : -1;
}
