// Tagger base
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

// Tagger base
#include "tagbase.h"
#include <cctype>
#include <limits>
#include <algorithm>
using namespace std;

//---------------------------------------------------------------------------
double TagBase::tinyScore = numeric_limits<double>::min();

string TagBase::anchorWord  = "00ANCHOR";
string TagBase::numberWord  = "00NUMBER";
string TagBase::unknownWord = "00UNKNOWN";

//---------------------------------------------------------------------------
bool TagBase::readBuffer(string &s, istream &in)
{
    if (!in.good())
    {
        return false;
    }
    getline(in, s, '\n');

    if (!in.good())
    {
        return false;
    }

    // With Windows and cygwin, we sometimes get a \r as well
    unsigned int slen = s.length();
    if (slen > 0 && s[slen-1] == '\r')
    {
        s = s.substr(0, slen-1);
    }

    return true;
}

bool TagBase::readWord(string &s, istream &in)
{
    if (!in.good())
    {
        return false;
    }
    in >> s;
    return in.good();
}

//---------------------------------------------------------------------------
// Number tests

// Code must match up with the ones given in the classes below; this is not
// terribly good coding practice, but it will do.
const TagBase::NumberTest *TagBase::getNumberTest(unsigned int code)
{
    return (code == 1) ? AllDigit
        : (code == 2) ? AnyDigit
        : (code == 3) ? ParseableNumber
        : (code == 4) ? NeverNumber
        : NULL;
}

struct AllDigitNumberTest : public TagBase::NumberTest
{
    bool isNumber(const string &s) const
    {
        string::size_type slen = s.length();
        for (string::size_type p = 0 ; p < slen ; ++p)
        {
            if (!isdigit(s[p]))
            {
                return false;
            }
        }
        return true;
    }

    unsigned int code() const
    {
        return 1;
    }
};
static AllDigitNumberTest all_digit;
const TagBase::NumberTest *TagBase::AllDigit = &all_digit;

struct AnyDigitNumberTest : public TagBase::NumberTest
{
    bool isNumber(const string &s) const
    {
        string::size_type slen = s.length();
        for (string::size_type p = 0 ; p < slen ; ++p)
        {
            if (isdigit(s[p]))
            {
                return true;
            }
        }
        return false;
    }

    unsigned int code() const
    {
        return 2;
    }
};
static AnyDigitNumberTest any_digit;
const TagBase::NumberTest *TagBase::AnyDigit = &any_digit;

struct ParsableNumberTest : public TagBase::NumberTest
{
    bool isNumber(const string &s) const
    {
        // This is a fairly weak test for parsable numbers:
        // optional '-' or '+' at start
        // Then any digits or commas
        // Then optional full stop
        // Then any digits
        // We don't actually check the commas separate the numbers in 3s, and
        // we don't require any digits before the point, though there has to
        // be some digit somewhere.
        string::size_type slen = s.length();
        string::size_type p = 0;
        bool any_digit = false;
        if (s[p] == '-' || s[p] == '+')
        {
            p += 1;
        }
        
        for (; p < slen; ++p)
        {
            if (isdigit(s[p]))
            {
                any_digit = true;
            }
            else if (s[p] == '.')
            {
                p += 1;
                break;
            }
            else if (s[p] != ',')
            {
                return false;
            }
        }

        // Now we are after the deciaml point
        for (; p < slen; ++p)
        {
            if (!isdigit(s[p]))
            {
                return false;
            }
            else
            {
                any_digit = true;
            }
        }

        return any_digit;
    }

    unsigned int code() const
    {
        return 3;
    }
};
static ParsableNumberTest parsable_test;
const TagBase::NumberTest *TagBase::ParseableNumber = &parsable_test;

struct NeverNumberTest : public TagBase::NumberTest
{
    bool isNumber(const string &) const
    {
        return false;
    }

    unsigned int code() const
    {
        return 4;
    }
};
static NeverNumberTest never_test;
const TagBase::NumberTest *TagBase::NeverNumber = &never_test;

//---------------------------------------------------------------------------
void TagBase::lowercase(string &s)
{
    string::size_type slen = s.length();
    for (string::size_type p = 0 ; p < slen ; ++p)
    {
        s[p] = tolower(s[p]);
    }
}

void TagBase::normalise(string &s, const NumberTest *test, bool casehold)
{
    // Never change special words
    if (s == anchorWord || s == numberWord || s == unknownWord)
    {
        return;
    }
    
    if (!test->isNumber(s) && !casehold)
    {
        lowercase(s);
    }
}

bool TagBase::number(string &s, const NumberTest *test)
{
    if (test->isNumber(s))
    {
        s = numberWord;
        return true;
    }
    else
    {
        return false;
    }
}

bool TagBase::casecmp
(const string &s1, const string &s2, string::size_type pos1, string::size_type pos2)
{
    string::size_type s1len = s1.length();
    string::size_type s2len = s1.length();

    for (; pos1 < s1len && pos2 < s2len ; ++pos1, ++pos2)
    {
        if (tolower(s1[pos1]) != tolower(s2[pos2]))
        {
            return false;
        }
    }

    return (pos1 == s1len && pos2 == s2len);
}

bool TagBase::ncasecmp
(const string &s1, const string &s2, string::size_type n,
 string::size_type pos1, string::size_type pos2)
{
    string::size_type s1len = s1.length();
    string::size_type s2len = s1.length();

    for (; n > 0 && pos1 < s1len && pos2 < s2len ; ++pos1, ++pos2, --n)
    {
        if (tolower(s1[pos1]) != tolower(s2[pos2]))
        {
            return false;
        }
    }
    
    return (n == 0 || (pos1 == s1len && pos2 == s2len));
}
