#ifndef TagBase_H
#define TagBase_H
// Tagger base
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

// Tagger base
#include <string>
#include <iostream>
#include <cfloat>
using namespace std;

//---------------------------------------------------------------------------
// The TagBase class is largely a holder for various types, constants and
//   utility functions used throughout the tagger.

class TagBase
{
    public:
    // A small value used to ensure some double values are non-zero.
    static double tinyScore;

    // Compare strings without case sensitivity.
    // Compare s1 starting at position pos1 with s2 starting at position pos2.
    // Return true if equal.
    static bool casecmp(const string &s1, const string &s2,
                 string::size_type pos1=0,
                 string::size_type pos2=0);

    // Compare strings without case sensitivity.
    // Compare s1 starting at position pos1 with s2 starting at position pos2.
    // Compare at most n characters.
    // Return true if equal.
    static bool ncasecmp(const string &s1, const string &s2,
                  string::size_type n,
                  string::size_type pos1=0,
                  string::size_type pos2=0);

    //------------------------------------------------------------------------
    // String I/O

    // Read from in to s, stopping at a newline (which is discarded). Also discards
    // \r which sometimes arises in an interaction between Windows and Cygwin.
    // Return true if the read succeeded.
    static bool readBuffer(string &s, istream &in);

    // Read from in to s, stopping at whitespace (which is not discarded).
    // Initial whitespace is skipped.
    // Return true if the read succeeded.
    static bool readWord(string &s, istream &in);

    // Skip to a newline
    static void readSkip(istream &in)
    {
        string dummy;
        getline(in, dummy, '\n');
    }

    //------------------------------------------------------------------------
    // General variables

    // These variables contain the names to use for special words. You can
    // change them, but this must be done before any training.

    // Name of the lexical entry placed between sentences.
    static string anchorWord;

    // Name of the lexical entry used to represent numbers.
    static string numberWord;

    // Name of the lexical entry used to hold unknown word properties.
    static string unknownWord;

    //------------------------------------------------------------------------
    // Number tests.

    // The Number Test class tests whether its argument should be treated as a
    // number. Each number test has a unique code, to help with I/O
    struct NumberTest
    {
        virtual bool isNumber(const string &s) const =0;
        virtual unsigned int code() const = 0;
    };

    // Get number test based on code.
    static const NumberTest *getNumberTest(unsigned int code);

    // Number if word consists only of digit characters
    static const NumberTest *AllDigit;

    // Number if word contains a digit character anywhere
    static const NumberTest *AnyDigit;

    // Number if word follows a rough pattern for numbers
    static const NumberTest *ParseableNumber;

    // Number test which never passes
    static const NumberTest *NeverNumber;

    //------------------------------------------------------------------------
    // Miscellaneous support functions

    // Lowercase a whole string in situ.
    static void lowercase(string &s);

    // Transform a token to the number word if it passes the number test.
    // Returns true if a conversion was made
    static bool number(string &s, const NumberTest *test);

    // General normaliser. This does a number check first and replaces the string
    // by the number word if it passes. Otherwise it lowercases the word if the
    // casehold flag is not set. This is normally the case only in training: in
    // tagging we can do a more sophisticated test by seeing what is in the lexicon.
    static void normalise(string &s, const NumberTest *test, bool casehold);
};

#endif // TagBase_H
