// Top level of tagger.
// Includes both main() and classes built on top of Bigram's classes, with the
// output methods added.
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "token.h"
#include "bigram.h"
#include "getopt.h"
#include <fstream>
using namespace std;

// Holder for evaluation data
struct EvalData
{
    EvalData() : total_words(0), total_ambig(0), correct_words(0), correct_ambig(0) {}
    unsigned int total_words;
    unsigned int total_ambig;
    unsigned int correct_words;
    unsigned int correct_ambig;
};

// Enumeration of output formats:
// Penn = similar to tagged Penn treebank.
// Flagged = vertical format with flags for ambig, unknown and incorrect.
// Untagged = no tags
enum OutputFormat {OutPenn, OutFlagged, OutUntagged};

// Output format and additional tags.
struct OutputFlags
{
    // Create default output format.
    OutputFlags() :
        output_format(OutPenn),
        all_tags(false), evaluate(false), give_scores(false),
        out_threshold(0) {}

    // Data members are public here as this is a local struct.
    // The overall format.
    OutputFormat output_format;

    // Flags for displaying all tags, accumulating evaluation measures, and
    // showing the scores.
    bool all_tags, evaluate, give_scores;

    // Threshold used to determine what we output.
    // If the threshold is negative, write entries until we have a total score
    // greater than or equal to -threshold.
    double out_threshold;
};

// Derived classes for tagging.
// These take an EvalData to hold information when evaluating.
class FB : public Bigram::FBTagger
{
    public:
    // Create FB tagger with a given output stream, evaluation block and
    // output flags.
    FB(ostream &o, EvalData &eval_, OutputFlags fl)
        : out(o), eval(eval_), format(fl) {}

    private:
    // Write the results
    void writer(bool initial) const;

    // Output stream
    ostream &out;

    // Evaluation block
    EvalData &eval;

    // Output flags
    OutputFlags format;
};

class Viterbi : public Bigram::ViterbiTagger
{
    public:
    // Create Viterbi tagger with a given output stream, evaluation block and
    // output flags.
    Viterbi(ostream &o, EvalData &eval_, OutputFlags fl)
        : out(o), eval(eval_), format(fl) {}

    private:
    // Write the results
    void writer(bool initial) const;

    // Output stream
    ostream &out;

    // Evaluation block
    EvalData &eval;

    // Output flags
    OutputFlags format;
};

// Accumulate evaluation from the node, adding to scores in eval.
// The flags are set to indicate whether we got a correct tag, whether the
// word was ambiguous, and whether it was unknown
static void evaluate
    (const Bigram::TagSpine &node, EvalData &eval,
     bool &correct, bool &ambig, bool &unknown)
{
    const TaggedToken &token = node.getToken();
    unsigned int ntags = node.tagCount();
    
    correct = false;
    ambig   = (ntags > 1);
    unknown = node.getLex()->isUnknown();

    eval.total_words += 1;
    if (ntags > 1)
    {
        TagSet::Tag *correct_tag;
        ambig = true;

        correct_tag = node.getTag(node.getChosen()).tag;
        eval.total_ambig += 1;

        for (unsigned int i = 0 ; i < token.tagCount() ; i++)
        {
            if (correct_tag == token.getTag(i).tag)
            {
                eval.correct_ambig += 1;
                eval.correct_words += 1;
                correct = true;
                break;
            }
        }
    }
    else
    {
        // Even if there is only one tag, we might have got it wrong for
        // unknown words
        if (node.getTag(node.getChosen()).tag == token.getTag(0).tag)
        {
            eval.correct_words += 1;
            correct = true;
        }
    }
}

// Se flags if not evaluating
static void set_flags
    (const Bigram::TagSpine &node, bool &correct, bool &ambig, bool &unknown)
{
    correct = true; // actually don't know
    ambig   = (node.tagCount() > 1);
    unknown = node.getLex()->isUnknown();
}

static void write_flags(ostream &out, bool correct, bool ambig, bool unknown)
{
    out << (ambig   ? 'a' : ' ');
    out << (unknown ? 'u' : ' ');
    out << (correct ? ' ' : '-');
    out << ' ';
}

void FB::writer(bool initial) const
{
    char tok_tag_sep = '/';
    char tag_tag_sep = '|';
    char tag_score_sep = '/';
    
    for (TagChart::const_iterator ch = chart.begin(); ch != chart.end(); ++ch)
    {
        if (ch == chart.begin() && !initial)
        {
            continue;
        }
        
        const Bigram::FBSpine &node = (const Bigram::FBSpine &)**ch;

        bool correct, ambig, unknown;
        if (format.output_format == OutFlagged)
        {
            set_flags(node, correct, ambig, unknown);
        }
    
        if (format.evaluate)
        {
            evaluate(node, eval, correct, ambig, unknown);
        }

        if (format.output_format == OutFlagged)
        {
            write_flags(out, correct, ambig, unknown);
        }

        // Write the token
        out << node.getToken().getToken();

        TagSet::TagScore *tags, tempTags;
        bool own_arrays = false;
        unsigned int ntags = 1;
        char sep = tok_tag_sep;

        if (format.all_tags)
        {
            // Make our own array and copy across.
            // We do an insertion sort on score as we do this. There are
            // usually only a few tags, so this is reasonable.
            tags = new TagSet::TagScore[ntags = node.tagCount()];
            own_arrays = true;

            for (unsigned int i = 0 ; i < ntags ; ++i)
            {
                const Lexicon::LexTag &t = node.getTag(i);
                double score = node.getData(i).gamma;
                bool found = false;
                for (unsigned int j = 0; j < i; ++j)
                {
                    if (score >= tags[j].score)
                    {
                        for (unsigned int k = i; k > j; --k)
                        {
                            tags[k] = tags[k-1];
                        }
                        tags[j].tag = t.tag;
                        tags[j].score = score;
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    tags[i].tag   = t.tag;
                    tags[i].score = score;
                }
            }
        }
        else
        {
            // Single tag: use temporary object
            unsigned int chosen = node.getChosen();
            tempTags.tag   = node.getTag(chosen).tag;
            tempTags.score = node.getData(chosen).gamma;
            tags = &tempTags;
        }

        // Write all entries with a score above the threshold
        bool wrote_any = false;
        if (format.out_threshold >= 0)
        {
            for (unsigned int i = 0;
                 i < ntags && tags[i].score >= format.out_threshold; ++i)
            {
                out << sep << tags[i].tag->getName();
                if (format.give_scores)
                {
                    out << tag_score_sep << tags[i].score;
                }
                sep = tag_tag_sep;
                wrote_any = true;
            }
        }
        else
        {
            double total = 0; // Actually negated total
            for (unsigned int i = 0; i < ntags && total >= format.out_threshold; ++i)
            {
                out << sep << tags[i].tag->getName();
                if (format.give_scores)
                {
                    out << tag_score_sep << tags[i].score;
                }
                sep = tag_tag_sep;
                total -= tags[i].score;
                wrote_any = true;
            }
        }

        // If we failed to write anything as a result of thresholding, then write
        // the topmost result now
        if (!wrote_any)
        {
            out << sep << tags[0].tag->getName();
            if (format.give_scores)
            {
                out << tag_score_sep << tags[0].score;
            }
        }

        if (own_arrays)
        {
            delete [] tags;
        }
    
        out << endl;
    }
}

void Viterbi::writer(bool initial) const
{
    char tok_tag_sep = '/';
    
    for (TagChart::const_iterator ch = chart.begin(); ch != chart.end(); ++ch)
    {
        if (ch == chart.begin() && !initial)
        {
            continue;
        }
        
        const Bigram::ViterbiSpine &node = (const Bigram::ViterbiSpine &)**ch;

        bool correct, ambig, unknown;
        if (format.output_format == OutFlagged)
        {
            set_flags(node, correct, ambig, unknown);
        }
    
        if (format.evaluate)
        {
            evaluate(node, eval, correct, ambig, unknown);
        }
    
        if (format.output_format == OutFlagged)
        {
            write_flags(out, correct, ambig, unknown);
        }
    
        // Write the token
        out << node.getToken().getToken();

        // Write the tag: ignore anything to do with score
        out <<  tok_tag_sep << node.getTag(node.getChosen()).tag->getName();
        out << endl;
    }
}

int main(int argc, char *argv[])
{
    int c, last_optind = 1;
    bool have_datafile = false;
    OutputFlags format;
    EvalData eval;
    bool evaluate = false;

    ifstream infile;
    ofstream outfile;
    istream *in  = NULL;
    ostream *out = &cout;

    enum Format { Penn, Tabbed, Untagged };
    Format input_format = Untagged;
    bool casehold = false;
    bool use_fb = true;

    Lexicon lexicon;
    TagSet  tagset;
    Endings rules;

    opterr = 0;

    // Verify the arguments
    while ((c = local_getopt(argc, argv, "m:i:f:c:o:vFaAsSt:TelL")) != -1)
    {
        switch (c)
        {
            case 'm':
                if (have_datafile)
                {
                    cerr << "Model already read: " << optarg
                        << " ignored" << endl;
                }
                else
                {
                    ifstream d(optarg, ios::in | ios::binary);
                    if (!d)
                    {
                        cerr << "Cannot open model file " << optarg << endl;
                    }
                    else
                    {
                        have_datafile = Bigram::readData
                            (d, tagset,lexicon, rules);
                    }
                }
                break;
            case 'i':
                if (strcmp(optarg, "p") == 0)
                    input_format = Penn;
                else if (strcmp(optarg, "t") == 0)
                    input_format = Tabbed;
                else if (strcmp(optarg, "u") == 0)
                    input_format = Untagged;
                else
                {
                    cerr << "Unknown input format '" << optarg
                        << "': using default" << endl;
                    input_format = Untagged;
                }
                break;
            case 'f':
                if (strcmp(optarg, "t") == 0)
                    format.output_format = OutPenn;
                else if (strcmp(optarg, "a") == 0)
                    format.output_format = OutFlagged;
                else
                {
                    cerr << "Unknown output format '" << optarg
                        << "': using default" << endl;
                    format.output_format = OutPenn;
                }
                break;
            case 'c':
                if (!have_datafile)
                {
                    cerr << "Cannot tag corpus: no datafile" << endl;
                }
                else
                {
                    format.evaluate = evaluate
                        && input_format != Untagged;

                    optind -= 1; // Reset to actual argument
                    
                    do
                    {
                        char *filename = argv[optind];
                        istream *new_in;
                        if (strcmp(filename, "-") == 0)
                        {
                            new_in = &cin;
                        }
                        else
                        {
                            infile.open(filename);
                            new_in = &infile;
                        }
                        
                        if (!new_in->good())
                        {
                            cerr << "Cannot open '" << filename
                                << "'" << endl;
                        }
                        else
                        {
                            if (in != NULL && in != &cin)
                            {
                                ((ifstream *)in)->close();
                            }
                            in = new_in;

                            // Create a tagger first, then run it
                            Bigram::Tagger *tagger = 
                                (use_fb)
                                ? (Bigram::Tagger *)(new FB(*out, eval, format))
                                : (Bigram::Tagger *)(new Viterbi(*out, eval, format));
                            TokenReader *reader
                                = (input_format == Penn)
                                ? TokenReader::reader_penn(tagset)
                                : (input_format == Tabbed)
                                ? TokenReader::reader_tabbed(tagset)
                                : (input_format == Untagged)
                                ? TokenReader::reader_untagged()
                                : NULL;

                            tagger->tagger(reader, lexicon, tagset, rules,
                                *in, casehold, false);
                            delete reader;
                            delete tagger;
                        }
                    } while (++optind < argc && argv[optind][0] != '-');
                }
                break;
            case 'o':
                {
                    ostream *new_out;
                    if (strcmp(optarg, "-") == 0)
                    {
                        new_out = &cout;
                    }
                    else
                    {
                        outfile.open(optarg);
                        new_out = &outfile;
                    }
                    
                    if (!out->good())
                    {
                        cerr << "Cannot open '" << optarg << "'" << endl;
                        new_out = &cout;
                    }
                    else
                    {
                        if (out != NULL && out != &cout)
                        {
                            ((ofstream *)out)->close();
                        }
                        out = new_out;
                    }
                    break;
                }
            case 'v':
                use_fb = false;
                break;
            case 'F':
                use_fb = true;
                break;
            case 'a':
                format.all_tags = true;
                break;
            case 'A':
                format.all_tags = false;
                break;
            case 's':
                format.give_scores = true;
                break;
            case 'S':
                format.give_scores = false;
                break;
            case 't':
                format.out_threshold = atof(optarg);
                break;
            case 'T':
                format.out_threshold = 0.0;
                break;
            case 'e':
                evaluate = true;
                break;
            case 'l':
                casehold = true;
                break;
            case 'L':
                casehold = false;
                break;
            default:
                cerr << "Bad or unknown option " << argv[optind-1] << endl;
                break;
        }

        last_optind = optind;
    }

    if (in != NULL  && in  != &cin)
    {
        ((ifstream *)in)->close();
    }
    
    if (out != NULL && out != &cout)
    {
        ((ofstream *)out)->close();
    }
    
    if (evaluate && eval.total_words > 0)
    {
        cerr << eval.total_words << " words ("
            << eval.correct_words << " correct), "
            << eval.total_ambig << " ambiguous ("
            << eval.correct_ambig << " correct)" << endl;
        cerr << "Success rate: "
            << ((double)100*eval.correct_words)/eval.total_words
            << "%, ambiguous "
            << ((eval.total_ambig > 0)
                ? ((double)100*eval.correct_ambig)/eval.total_ambig : 0)
            << "%" << endl;
    }

    return 0;
}
