// Tag set and transitions
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagset.h"
#include <limits>
#include <cctype>
#include <list>
using namespace std;

TagSet::Tag *TagSet::NoTag = NULL;
unsigned int NoIndex = numeric_limits<unsigned int>::max();

void TagSet::clear()
{
    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        delete (*i).second;
    }
    tags.clear();
    anchorTag = NoTag;
}

bool TagSet::read(istream &in)
{
    bool ok = true;
    bool have_anchor = false;
    string buffer;

    while (TagBase::readBuffer(buffer, in))
    {
        // Look for a space followed by flags.
        string::size_type end = buffer.find(' ');
        string name;
        char flag;

        if (end != buffer.npos)
        {
            name = buffer.substr(0, end);
            while (end < buffer.length() && isspace(buffer[end]))
            {
                ++end;
            }
            flag = (end == buffer.length()) ? '\0' : tolower(buffer[end]);
        }
        else
        {
            name = buffer;
            flag = '\0';
        }

        // Check for empty name
        if (name.length() == 0)
        {
            cerr << "Empty tag name" << endl;
            ok = false;
        }
        else
        {
            // Check for duplicates, verify properties, then create the tag
            if (tags.find(name) != tags.end())
            {
                cerr << "Duplicate tag " << name << endl;
                ok = false;
            }

            if (flag != 'c' && flag != 'a' && flag != '\0')
            {
                cerr << "Bad tag flag " << flag << " at " << name << endl;
                ok = false;
            }
            
            bool closed = (flag == 'c' || flag == 'a');
            bool anchor = (flag == 'a');
            if (anchor && have_anchor)
            {
                cerr << "Can only have one anchor tag at " << name << endl;
                ok = false;
            }

            if (ok)
            {
                tags[name] = makeTag(name, closed, anchor);
            }
        }
    }

    // Set up transitions and spot the anchor
    unsigned int ntags = tags.size();
    unsigned int number = 0;
    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        Tag *tag = (*i).second;
        if (tag->anchor)
        {
            anchorTag = tag;
        }

        tag->number = number++;
        tag->set(ntags);
    }

    return ok;
}

// Write tags list and transitions into a file (binary/text)
/* Binary format:
   <ntags>
   [<flags><tag><total><score array>]+

   Text format:
   <TAGS>
   <tag><flags><total>:<scores>
   </TAGS>
   flags is 'c', 'a' or '.'
 */
void TagSet::write(ostream &out, bool textual) const
{
    unsigned int ntags = tags.size();
    if (textual)
    {
        out << "<TAGS>" << endl;
        for (TagMap::const_iterator i = tags.begin(); i != tags.end(); ++i)
        {
            const Tag *tag = (*i).second;
            out << tag->name << '\t';
            out << ((tag->anchor) ? 'a' : ((tag->closed) ? 'c' : '.'));
            out << '\t' << tag->total << " : ";

            for (Trans::const_iterator j = tag->trans.begin();
                 j != tag->trans.end(); ++j)
            {
                out << ' ' << *j;
            }
            out << endl;
        }
        out << "</TAGS>" << endl;
    }
    else
    {
        out.write((char *)&ntags, sizeof(ntags));
        for (TagMap::const_iterator i = tags.begin(); i != tags.end(); ++i)
        {
            const Tag *tag = (*i).second;
            out.write((char *)&(tag->closed), sizeof(tag->closed));
            out.write((char *)&(tag->anchor), sizeof(tag->anchor));
            out << tag->name << '\n';
            out.write((char *)&(tag->total), sizeof(tag->total));
            
            for (Trans::const_iterator j = tag->trans.begin();
                 j != tag->trans.end(); ++j)
            {
                double score = *j;
                out.write((char *)&score, sizeof(score));
            }
        }
    }
}

void TagSet::writeTrans(ostream &out, bool textual) const
{
    for (TagMap::const_iterator i = tags.begin(); i != tags.end(); ++i)
    {
        const Tag *itag = (*i).second;

        for (TagMap::const_iterator j = tags.begin(); j != tags.end(); ++j)
        {
            const Tag *jtag = (*j).second;
            unsigned int jnum = jtag->number;
            
            out << itag->name << ' ';
            out << jtag->name << ' ';
            if (textual)
            {
                out << itag->trans[jnum];
            }
            else
            {
                out.write((char *)&(itag->trans[jnum]), sizeof(itag->trans[jnum]));
            }
            out << endl;
        }
    }
}

static string read_err = "Tagset read failed";
bool TagSet::readBinary(istream &in)
{
    unsigned int ntags;

    in.read((char *)&ntags, sizeof(ntags));
    if (!in.good())
    {
        cerr << read_err << endl;
        return false;
    }

    string anchor_name;

    unsigned int number = 0;
    for (unsigned int i = 0 ; i < ntags ; ++i)
    {
        bool closed;
        bool anchor;
        string name;

        in.read((char *)&closed, sizeof(closed));
        in.read((char *)&anchor, sizeof(anchor));

        if (!in.good())
        {
            cerr << read_err << endl;
            return false;
        }
        
        if (!TagBase::readBuffer(name, in))
        {
            cerr << read_err << endl;
            return false;
        }

        Tag *tag = tags[name] = makeTag(name, closed, anchor);

        // Record the anchor tag when we see it
        if (anchor)
        {
            anchor_name = name;
        }
        
        in.read((char *)&(tag->total), sizeof(tag->total));
        if (!in.good())
        {
            cerr << read_err << endl;
            return false;
        }
        
        tag->set(ntags);
        for (unsigned int j = 0; j < ntags; ++j)
        {
            double transition;
            in.read((char *)&transition, sizeof(transition));
            if (!in.good())
            {
                cerr << read_err << endl;
                return false;
            }
            tag->trans[j] = transition;
        }    
        tag->number = number++;
    }
    
    anchorTag = tags[anchor_name];
    return true;
}

void TagSet::transNormalise()
{
    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        Tag *tag = (*i).second;
        double norm = tag->total;

        if (norm > TagBase::tinyScore)
        {
            for (Trans::iterator j = tag->trans.begin(); j != tag->trans.end(); ++j)
            {
                *j = (*j / norm) + TagBase::tinyScore;
            }
        }
        else
        {
            tag->trans.assign(tags.size(), TagBase::tinyScore);
        }
    }
}

void TagSet::transDenormalise()
{
    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        Tag *tag = (*i).second;
        double norm = tag->total;

        for (Trans::iterator j = tag->trans.begin(); j != tag->trans.end(); ++j)
        {
            *j *= norm;
        }
    }
}

TagSet::Tag *TagSet::find(const string &tag) const
{
    TagMap::const_iterator at = tags.find(tag);
    return (at == tags.end()) ? NoTag : (*at).second;
}

bool TagSet::mergeIn(const TagSet &tagset)
{
    // If current tagset is empty, just transfer the data
    // Spot the anchor tag as we do it
    if (tags.size() == 0)
    {
        for (TagMap::const_iterator i = tagset.tags.begin(); i != tagset.tags.end(); ++i)
        {
            const string &name = (*i).first;
            const Tag *from    = (*i).second;

            // Duplicate the tag
            Tag *newTag   = tags[name] = makeTag(name, from->closed, from->anchor);
            newTag->total = from->total;

            unsigned int ntags = from->trans.size();
            newTag->set(ntags);

            for (unsigned int i = 0 ; i < ntags ; ++i)
            {
                newTag->trans[i] = from->trans[i];
            }
    
            newTag->number = from->number;

            if (from->anchor)
            {
                anchorTag = newTag;
            }
        }
        return true;
    }

    // Check tagsets for consistency, and merge if ok
    unsigned int ntags = tags.size();
    bool ok = (ntags == tagset.tags.size());
    for (TagMap::const_iterator i = tagset.tags.begin() ;
         ok && i != tagset.tags.end(); ++i)
    {
        const string &name = (*i).first;
        const Tag    *otag = (*i).second;

        Tag *tag = find(name);
        if (tag != NULL
            && tag->closed == otag->closed
            && tag->anchor == otag->anchor)
        {
            if (otag->total != TagBase::tinyScore)
            {
                tag->total += otag->total;

                for (unsigned  int j = 0 ; j < ntags ; ++j)
                {
                    if (otag->trans[j] != TagBase::tinyScore)
                    {
                        tag->trans[j] += otag->trans[j];
                    }
                }
            }
        }
        else
        {
            ok = false;
        }
    }

    if (!ok)
    {
        cerr << "Tagsets are not consistent" << endl;
    }
    return ok;
}

void TagSet::openTags(set<Tag*> &result) const
{
    for (TagMap::const_iterator i = tags.begin(); i != tags.end(); ++i)
    {
        const Tag *tag = (*i).second;
        if (!tag->closed)
        {
            result.insert((Tag *)tag);
        }
    }
}

void TagSet::allTags(set<Tag*> &result) const
{
    for (TagMap::const_iterator i = tags.begin(); i != tags.end(); ++i)
    {
        const Tag *tag = (*i).second;
        result.insert((Tag *)tag);
    }
}

//-----------------------------------------------------------------------------------
void ReestTagSet::ReestTag::commit(unsigned int ntags)
{
    for (unsigned int j = 0; j < ntags; ++j)
    {
        trans[j] = rtrans[j] + TagBase::tinyScore;
    }
    total = rtotal + TagBase::tinyScore;
}

void ReestTagSet::ReestTag::reset(unsigned int ntags)
{
    rtrans.assign(ntags, 0);
    rtotal = 0.0;
}

void ReestTagSet::ReestTag::addTo(const Tag *to, double score, double beta)
{
    rtrans[to->getNumber()] += score;
    rtotal += beta;
}

void ReestTagSet::commit()
{
    unsigned int ntags = tags.size();

    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        ReestTag *tag = (ReestTag *)((*i).second);
        tag->commit(ntags);
    }
}

void ReestTagSet::reset()
{
    unsigned int ntags = tags.size();

    for (TagMap::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        ReestTag *tag = (ReestTag *)((*i).second);
        tag->reset(ntags);
    }
}

