#ifndef TagSet_H
#define TagSet_H
// Tag set and transitions
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagbase.h"
#include <vector>
#include <map>
#include <set>
using namespace std;

// The TagSet hold information about all valid tags.
// In includes the array of bigram transition scores.
// A second class which includes space for re-estimation follows.
class TagSet
{
    public:
    typedef vector<double> Trans;
    
    // Representation of an individual tag.
    // Each tag has a name, number flags, total score and array of transitions,
    // indexed by tag number.
    class Tag
    {
        friend class TagSet;
        
        public:
        // Construct with name and flags, setting number to zero for now
        Tag(const string &name_ = "", bool closed_ = false, bool anchor_ = false)
            : name(name_), closed(closed_), anchor(anchor_),
              total(TagBase::tinyScore), number(0)
        {}

        // Destroy a tag
        virtual ~Tag() { trans.clear(); }

        // Set up transitions array to the given size.
        virtual void set(unsigned int ntags)
        {
            trans.assign(ntags, TagBase::tinyScore);
        }

        // Find the transition score between two tags
        double tagToTag(const Tag *to) const
        {
            return trans[to->number];
        }

        // Add to a given transition, keeping the total up to date.
        void addTagToTag(const Tag *to, double score)
        {
            total += score;
            trans[to->number] += score;
        }

        // Replace a given transition, keeping the total up to date.
        void replaceTagToTag(const Tag *to, double score)
        {
            total -= trans[to->number];
                
            total += score;
            trans[to->number] = score;
        }

        // Get the name.
        const string &getName() const
        {
            return name;
        }

        // Get the number (index).
        unsigned int getNumber() const
        {
            return number;
        }

        // Get the transition total
        double getTotal() const
        {
            return total;
        }

        // Add to the total (mainly for fixing unknown value)
        void addToTotal(double increment)
        {
            total += increment;
        }

        protected:
        // Name of the tag
        string name;

        // Flags for closed and anchor tags
        bool closed, anchor;

        // Total of the transitions
        double total;

        // Transitions array
        Trans trans;

        // Number of the tag
        unsigned int number;
    };

    // A tag with a score
    struct TagScore
    {
        TagScore() : tag(NULL), score(0) {}
        TagScore(Tag *t, double s) : tag(t), score(s) {}
        
        Tag   *tag;
        double score;
    };

    // Special value indicating an unknown or undefined tag.
    static Tag *NoTag;

    // Special value indicating index of an unknown or undefined tag.
    static unsigned int NoIndex;

    protected:
    // Overall tagset representation: map names to information.
    // Note: pointers to tags only remain stable if no more tags are added.
    typedef map<string, Tag *> TagMap;
    TagMap tags;

    // The anchor tag
    Tag *anchorTag;

    public:
    // Create empty tagset.
    TagSet() : anchorTag(NULL) {}

    // Free a tagset
    virtual ~TagSet() { clear(); }

    // Get the anchor tag
    Tag *getAnchor() const { return anchorTag; }

    // Clear the tagset. Tag points will be invalidated.
    void clear();

    // Read tags list from a file and create a new tagset (not cleared)
    bool read(istream &in);

    // Write tags list and transitions into a file (binary or text)
    void write(ostream &out, bool textual) const;

    // Write just the transitions (binary or text)
    void writeTrans(ostream &out, bool textual = true) const;

    // Read tags list and transitions (binary). Tagset is not clear first.
    bool readBinary(istream &in);

    // Normalise scores in transitions
    void transNormalise();

    // Denormalise: requires the totals to be correct
    void transDenormalise();

    // Find a tag in a tagset. Returns NoTag if not known.
    Tag *find(const string &tag) const;

    // Find the number of a tag in the tagset, or NoIndex if non existent
    unsigned int findIndex(const string &tag) const
    {
        Tag *at = find(tag);
        return (at == NULL) ? NoIndex : at->number;
    }

    // Merge another tagset into this one (both must be denormalised).
    // Report any inconsistencies on cerr.
    // Return false if the merge failed.
    bool mergeIn(const TagSet &tagset);

    // Get a set of all tags which are open.
    void openTags(set<Tag *> &result) const;

    // Get a set of all tags.
    void allTags(set<Tag *> &result) const;

    // Report how many tags there are
    unsigned int tagCount() const
    {
        return tags.size();
    }

    protected:
    // Make a tag
    virtual Tag *makeTag(const string &name, bool closed, bool anchor)
    {
        return new Tag(name, closed, anchor);
    }
};

// Variant of tagset which includes space for re-estimation
class ReestTagSet : public TagSet
{
    public:
    // Extension of Tag used when we are re-estimating
    class ReestTag : public TagSet::Tag
    {
        public:
        // Construct with name and flags, setting number to zero for now
        ReestTag
        (const string &name = "", bool closed = false, bool anchor = false)
            : TagSet::Tag(name, closed, anchor)
        {}

        // Destroy reest tag
        ~ReestTag() { rtrans.clear(); }

        // Set up transitions array to the given size.
        void set(unsigned int ntags)
        {
            rtrans.assign(ntags, TagBase::tinyScore);
            Tag::set(ntags);
        }

        // Commit the result from re-estimation to main transitions.
        // Do before normalisation.
        void commit(unsigned int ntags);

        // Reset re-estimation values
        void reset(unsigned int ntags);

        // Add score to the transition and beta to the total for the transition
        // to indicated tag.
        void addTo(const Tag *to, double score, double beta);

        private:
        // Total of re-estimation transitions
        double rtotal;

        // The re-estimation transitions
        Trans rtrans;
    };

    // Create empty tagset.
    ReestTagSet() {}

    // Free a tagset
    ~ReestTagSet() {}

    // Commit the result from re-estimation to main transitions.
    // Do before normalisation.
    void commit();

    // Reset re-estimation values
    void reset();

    protected:
    // Make a tag
    Tag *makeTag(const string &name, bool closed, bool anchor)
    {
        return new ReestTag(name, closed, anchor);
    }
};

#endif // TagSet_H
