// Token reader
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "token.h"
#include <cctype>
using namespace std;

TaggedToken TaggedToken::blankToken;

void TaggedToken::clear()
{
    token = "";
    tags.clear();
}

void TaggedToken::anchor(const TagSet &tagset)
{
    token = TagBase::anchorWord;
    tags.clear();
    tags.push_back(TagSet::TagScore(tagset.getAnchor(), 1));
}


//---------------------------------------------------------------------------
// Standard formats and reader.
static SepTagTokenReader::SepTagTokenControl
    pennFormat('/', '|', '\\', "*x*", true, true);

static SepTagTokenReader::SepTagTokenControl
    tabbedFormat('\t', '-', '\0', "<", false, true);

TokenReader *TokenReader::reader_penn(const TagSet &tagset)
{
    return new SepTagTokenReader(tagset, pennFormat);
}

TokenReader *TokenReader::reader_tabbed(const TagSet &tagset)
{
    return new SepTagTokenReader(tagset, tabbedFormat);
}

TokenReader *TokenReader::reader_untagged()
{
    return new UntaggedTokenReader();
}


SepTagTokenReader::SepTagTokenReader
    (const TagSet &tagset_, SepTagTokenControl &control_)
    : tagset(tagset_), control(control_)
{
}

bool SepTagTokenReader::read(istream &in, TaggedToken &dest) const
{
    dest.clear();

    // Read until we get a tagged word
    string buffer;
    string::size_type skip_len = control.skip.length();

    do
    {
        bool ok = true;
        bool any_read = (control.word_based)
            ? TagBase::readWord(buffer, in)
            : TagBase::readBuffer(buffer, in);

        if (!any_read)
        {
            return false;
        }
        
        // Check for line skip
        if (skip_len != 0 && buffer.compare(0, skip_len, control.skip) == 0)
        {
            // Skip rest of line if word based.
            if (control.word_based)
            {
                TagBase::readSkip(in);
            }

            // Add an anchor if needed.
            if (control.anchor_on_skip)
            {
                dest.anchor(tagset);
                return true;
            }
            else
            {
                continue;
            }
        }

        // Look to see if we have a tag; check in case escaped
        string::size_type tag_at = 0;
        do
        {
            tag_at = buffer.find(control.wordsep, tag_at+1);
        } while (tag_at != buffer.npos && buffer[tag_at-1] == control.escape);
        
        if (tag_at != buffer.npos)
        {
            double tag_score;
            string::size_type tag_start = tag_at;

            // Count and verify the tags, with an optimisation for the case
            // where there is just one tag
            if (buffer.find(control.tagsep, tag_at+1) == buffer.npos)
            {
                // Single tag
                TagSet::Tag *tag = tagset.find(buffer.substr(tag_at+1));
                if (tag == NULL)
                {
                    cerr << "Unknown tag " << buffer.substr(tag_at+1)
                        << " in " << buffer << endl;
                    ok = false;
                }
                else
                {
                    dest.clear();
                    dest.addTag(tag, 1);
                }
            }
            else
            {
                // There are multiple tags. Count them first.
                unsigned int ntags = 0;
                for (string::size_type t = tag_at; t != buffer.npos; ++ntags)
                {
                    t = buffer.find(control.tagsep, t+1);
                }

                tag_score = ((double)1)/ntags;
                dest.clear();

                while (tag_at != buffer.npos)
                {
                    string::size_type tag_end
                        = buffer.find(control.tagsep, ++tag_at);
                    string tag_name = buffer.substr(tag_at, tag_end-tag_at);
                    TagSet::Tag *tag = tagset.find(tag_name);
                    if (tag == NULL)
                    {
                        cerr << "Unknown tag " << tag_name << endl;
                        ok = false;
                    }
                    dest.addTag(tag, tag_score);
                    tag_at = tag_end;
                }
            }

            if (dest.tagCount() != 0 && ok)
            {
                dest.setToken(buffer.substr(0, tag_start));
                return true;
            }
            else
            {
                cerr << "Tags missing at " << buffer << endl;
            }
        }
    } while (true);
}

bool UntaggedTokenReader::read(istream &in, TaggedToken &dest) const
{
    dest.clear();

    string token;
    if (TagBase::readWord(token, in))
    {
        dest.setToken(token);
        return true;
    }
    else
    {
        return false;
    }
}
