#ifndef Token_H
#define Token_H
// Token reader
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagset.h"
using namespace std;

/* Tokens are read into a structure consisting of a word, and an array of tags
   with scores. We provide a general function type for reading such structures
   Even when we are dealing with untagged corpora, we will use this structure,
   so we don't have to worry about different formats.
*/

class TaggedToken
{
    public:
    // Create a tagged token
    TaggedToken() {}

    // Destroy a tagged token
    ~TaggedToken() {}

    // A token with null fields
    static TaggedToken blankToken;

    // Clear a tagged token structure
    void clear();

    // Get number of tags
    unsigned int tagCount() const
    {
        return tags.size();
    }

    // Set up this token as an anchor word.
    // Assumes exactly one anchor tag.
    void anchor(const TagSet &tagset);

    // Set the token.
    void setToken(const string &s)
    {
        token = s;
    }
    
    // Get the token.
    const string &getToken() const
    {
        return token;
    }
    
    // Get the token, modifiable version.
    string &getToken()
    {
        return token;
    }
    
    // Add a tag with a given score.
    void addTag(TagSet::Tag *tag, double score)
    {
        tags.push_back(TagSet::TagScore(tag, score));
    }

    // Get the ith tag.
    const TagSet::TagScore &getTag(unsigned int i) const
    {
        return tags[i];
    }

    private:
    typedef vector<TagSet::TagScore> Tags;

    // The token itself
    // Main token representation
    string token;

    // The list of tags
    Tags tags;
};

// For reading tokens, we use classes derived from TokenReader
class TokenReader
{
    public:
    // Create a token reader.
    TokenReader() {}

    // Destroy a token reader
    virtual ~TokenReader() {}

    // The basic read function. Caller should clear the token first.
    // Returns false if no token was read.
    virtual bool read(istream &in, TaggedToken &dest) const =0;

    // Static methods to create token reader for various standard formats
    // In each case except untagged, a tagset is used.
    // Decisions about what to recognize as number are made by the number
    // field of the lexicon; decisions about case are made after tokenization.
    static TokenReader *reader_penn(const TagSet &tagset);
    static TokenReader *reader_tabbed(const TagSet &tagset);
    static TokenReader *reader_untagged();
};

// Penn/LOB style reader.
// Word forma is: <word><sep>[<tag><tagsep>]*<tag>
// Untagged words are dropped
// If a word is seen to start with the skip string, the rest of the line is skipped.
// Can also be used for tab separated format.
// Flag tells us whether the usual read unit is a white space separated word,
// or a whole line.
// Second flag says what to do with a skip line: really just skip it, or fake
// an anchor at this point (e.g. if skip is a sentence identifier).
class SepTagTokenReader : public TokenReader
{
    public:
    // Internal structure which holds control information
    struct SepTagTokenControl
    {
        // Create with given word separator, tag separator, separator escape character,
        // skip string, and flags for whether we read a word or a line at a time and
        // whether we skip on an anchor.
        SepTagTokenControl
            (char wordsep_, char tagsep_, char escape_,
             const string &skip_, bool by_word = true, bool anchor_skip = false)
            : wordsep(wordsep_), tagsep(tagsep_), escape(escape_),
              skip(skip_), word_based(by_word), anchor_on_skip(anchor_skip)
        {}

        char wordsep, tagsep, escape;
        string skip;
        bool word_based, anchor_on_skip;
    };

    // Create a reader with a given tagset.
    // Normally, you will use one of the static methods in TokenReader, but we
    // make the class visible so additional formats can be built on it.
    SepTagTokenReader(const TagSet &tagset, SepTagTokenControl &control);

    // Read a token.
    bool read(istream &in, TaggedToken &dest) const;

    protected:
    const TagSet &tagset;
    SepTagTokenControl &control;
};

// Reader for untagged tokens.
class UntaggedTokenReader : public TokenReader
{
    public:
    // Create a reader.
    // It simply reads a word at a time, terminated by white space.
    UntaggedTokenReader() {}

    // Read a token.
    bool read(istream &in, TaggedToken &dest) const;
};

#endif // Token_H
