// Bigram model training
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "train.h"
#include "bigram.h"
#include "sgt.h"

//------------------------------------------------------------------------------------
// Increment correction class.
class IncrementCorrection : public Trainer::Correction
{
    public:
    // Create a correction object.
    IncrementCorrection(double inc) : increment(inc) {}
    
    // Destroy a correction object.
    ~IncrementCorrection() {}

    // Apply corrections to transitions (in tagset)
    void correctTagset(TagSet &tagset);

    private:
    // The increment
    double increment;
};

void IncrementCorrection::correctTagset(TagSet &tagset)
{
    set<TagSet::Tag *> tags;
    tagset.allTags(tags);
    
    for (set<TagSet::Tag *>::iterator i = tags.begin(); i != tags.end(); ++i)
    {
        TagSet::Tag *tag = *i;
        for (set<TagSet::Tag *>::iterator j = tags.begin(); j != tags.end(); ++j)
        {
            tag->addTagToTag(*j, increment);
        }
    }
}

// Simple Good-Turning correction class.
class SGTCorrection : public Trainer::Correction
{
    public:
    // Create a correction object.
    SGTCorrection() {}
    
    // Destroy a correction object.
    ~SGTCorrection() {}

    // Apply corrections to transitions (in tagset)
    void correctTagset(TagSet &tagset);
};

void SGTCorrection::correctTagset(TagSet &tagset)
{
    SGT<double> sgt;

    // Transfer all transitions data to the sgt object.
    // Each transition is a data point.
    set<TagSet::Tag *> tags;
    set<TagSet::Tag *>::iterator i;
    tagset.allTags(tags);
    
    for (i = tags.begin(); i != tags.end(); ++i)
    {
        TagSet::Tag *tag = *i;
        for (set<TagSet::Tag *>::iterator j = tags.begin(); j != tags.end(); ++j)
        {
            sgt.add(tag->tagToTag(*j), 1);
        }
    }

    sgt.analyse();

    // Now get the probability estimated back and scale them back to the original space
    double total = sgt.total();
    for (i = tags.begin(); i != tags.end(); ++i)
    {
        TagSet::Tag *tag = *i;
        for (set<TagSet::Tag *>::iterator j = tags.begin(); j != tags.end(); ++j)
        {
            double trans = tag->tagToTag(*j);
            double est;
            sgt.estimate(trans, est);
            tag->replaceTagToTag(*j, est * total);
        }
    }
}


Trainer::Correction *Trainer::makeCorrectionNone()
{
    return new Correction;
}

Trainer::Correction *Trainer::makeCorrectionIncrement(double increment)
{
    return new IncrementCorrection(increment);
}

Trainer::Correction *Trainer::makeCorrectionSGT()
{
    return new SGTCorrection();
}

//------------------------------------------------------------------------------------
bool Trainer::start(istream &tagset_file, Lexicon &lexicon)
{
    lex = &lexicon;
    return tagset.read(tagset_file);
}

void Trainer::train
(istream &corpus, bool echo, TokenReader *reader, bool casehold)
{
    TaggedToken token1, token2;
    TaggedToken *left = &token1, *right = &token2;
    Lexicon::LexEntry *left_lex, *right_lex;
    bool more = true;

    // Create an initial anchor word
    left->anchor(tagset);
    left_lex = lex->findOrCreate(left->getToken());

    unsigned int left_ntags = left->tagCount();
    for (unsigned int l = 0 ; l < left_ntags ; ++l)
    {
        left_lex->addTag(left->getTag(l));
    }

    // Process the rest of the corpus, adding a final anchor
    while (more)
    {
        // Read a token, preserving case, for now
        if (reader->read(corpus, *right))
        {
            if (echo)
            {
                cerr << right->getToken() << endl;
            }
            TagBase::normalise(right->getToken(), lex->getNumberTest(), casehold);
            if (echo)
            {
                cerr << "==> " << right->getToken() << endl;
            }
        }
        else
        {
            more = false;
            right->anchor(tagset);
        }

        right_lex = lex->findOrCreate(right->getToken());

        // Save lexical tags
        unsigned int right_ntags = right->tagCount();
        for (unsigned int r = 0 ; r < right_ntags ; ++r)
        {
            const TagSet::TagScore &rtag = right->getTag(r);
            right_lex->addTag(rtag);

            // Save transitions
            unsigned int left_ntags = left->tagCount();
            for (unsigned int l = 0 ; l < left_ntags ; ++l)
            {
                const TagSet::TagScore &ltag = left->getTag(l);
                ltag.tag->addTagToTag(rtag.tag, ltag.score * rtag.score);
            }
        }

        swap(left, right);
        swap(left_lex, right_lex);
    }
}

void Trainer::end(Correction *correction, const Endings &rules, ostream &out,
           double threshold, bool textual, bool no_norm )
{
    if (!textual)
    {
        no_norm = false;
    }

    // Ensure we have an unknown word.
    lex->setUnknown(tagset, 1);

    // Ensure anchor is set up.
    lex->adjustAnchor(tagset);

    correction->correctTagset(tagset);
    
    if (!no_norm)
    {
        tagset.transNormalise();
        lex->normalise();
    }

    if (threshold > 0)
    {
        lex->reduce(threshold);
    }
    
    Bigram::writeData(out, tagset, *lex, rules, textual);
}
