#ifndef Train_H
#define Train_H
// Bigram model training
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "tagset.h"
#include "lexicon.h"
#include "token.h"
#include "endings.h"

class Trainer
{
    public:
    // Correction class. This is used to adjust counts after training.
    // Corrections apply only to the transitions.
    // The base implementation does nothing.
    class Correction
    {
        public:
        // Create a correction object.
        Correction() {}

        // Destroy a correction object.
        virtual ~Correction() {}

        // Apply corrections to transitions (in tagset)
        virtual void correctTagset(TagSet &) {}
    };
    
    // Static methods to make some standard correction objects.
    // None changes nothing; Increment adds a constant to counts;
    // SGT applies Good-Turing.
    static Correction *makeCorrectionNone();
    static Correction *makeCorrectionIncrement(double increment = 1.0);
    static Correction *makeCorrectionSGT();

    // Create a trainer over a given tagset
    Trainer(TagSet &t) : tagset(t), lex(NULL) {}

    // Destroy trainer
    ~Trainer() {}

    // Carry out initialisation prior to training. It reads the (textual) tagset file
    // and stores a pointer to the lexicon.
    // Returns false if the tagset file could not be read.
    bool start(istream &tagset_file, Lexicon &lexicon);

    // Set the lexicon pointer - an alternative to start.
    void setLexicon(Lexicon *lexicon) { lex = lexicon; }

    // Get a reference to the lexicon
    Lexicon &getLexicon() const { return *lex; }

    // Train a bigram model from a corpus.
    // This augments an existing model, if there is one.
    // casehold gives the policy for lowercasing; echo says whether to echo
    // the input.
    void train(istream &corpus, bool echo, TokenReader *reader, bool casehold);

    // Carry out tidying up at end of training: creates unknown word, corrects
    // and normalises values, and writes file.
    // A flag says whether to use tetxual or binary mode for output.
    // You can also specify no normalisation, but only if textual is set.
    // Also does reduction on frequency if threshold > 0.
    // correction may not be null; use CorrectionNone if you want no correction.
    void end(Correction *correction, const Endings &rules, ostream &out,
        double threshold, bool textual = false, bool no_norm = false);

    private:
    // Tagset and transitions.
    TagSet  &tagset;

    // Lexicon: the pointer as we set it on the train call.
    Lexicon *lex;
};

#endif // Train_H
