// Bigram model training application
//
// Copyright (C) David Elworthy 2004
// All rights reserved.
//
// Redistribution and use in source and binary forms for any purpose, with or
// without modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions, and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions, and the disclaimer that follows 
//    these conditions in the documentation and/or other materials 
//    provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
// NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You may contact me at david@friendlymoose.com.

#include "train.h"
#include "getopt.h"
#include <fstream>
#include <cstdlib>
using namespace std;

int main(int argc, char *argv[])
{
    enum Format { Penn, Tabbed, Unknown, NotGiven };
    Format format = NotGiven;
    const TagBase::NumberTest *number = TagBase::AnyDigit;
    
    string taglist, outname, endname;
    
    ifstream tagfile, endfile;
    ofstream outfile;

    int c;
    bool ok = true, echo = false, casehold = false, report = false;
    Trainer::Correction *corr = Trainer::makeCorrectionNone();

    bool textual     = false;
    bool reduce_lex  = false;
    unsigned int reduce_level = 0;
    
    opterr = 0;

    // Get the arguments
    while ((c = local_getopt(argc, argv, "t:e:o:i:nNgGr:clFT" )) != -1)
    {
        switch (c)
        {
            case 't':
                if (taglist != "")
                {
                    cerr << "Tag list specified more than once" << endl;
                    ok = false;
                }
                else
                {
                    taglist = optarg;
                }
                break;
            case 'e':
                if (endname != "")
                {
                    cerr << "Endings rules specified more than once" << endl;
                    ok = false;
                }
                else
                {
                    endname = optarg;
                }
                break;
            case 'o':
                if (outname != "")
                {
                    cerr << "Output file specified more than once" << endl;
                    ok = false;
                }
                else
                {
                    outname = optarg;
                }
                break;
            case 'i':
                if (format != NotGiven)
                {
                    cerr << "Format specified more than once" << endl;
                    ok = false;
                }
                else if (strcmp(optarg, "p") == 0)
                {
                    format = Penn;
                }
                else if (strcmp(optarg, "t") == 0)
                {
                    format = Tabbed;
                }
                else
                {
                    cerr << "Unknown format" << endl;
                    ok = false;
                    format = Unknown;
                }
                break;
            case 'n':
                number = TagBase::AllDigit;
                break;
            case 'N':
                number = TagBase::ParseableNumber;
                break;
            case 'g':
                corr = Trainer::makeCorrectionSGT();
                break;
            case 'G':
                corr = Trainer::makeCorrectionIncrement();
                break;
            case 'c':
                echo = true;
                break;
            case 'l':
                casehold = true;
                break;
            case 'F':
                report = true;
                break;
            case 'T':
                textual = true;
                break;
            case 'r':
                reduce_lex = true;
                reduce_level = atoi(optarg);
                break;
            case '?':
                cerr << "Bad or unknown option " << argv[optind-1] << endl;
                break;
        }
    }

    if (taglist == "")
    {
        cerr << "A tag list must be specified" << endl;
        ok = false;
    }
    
    if (outname == "")
    {
        cerr << "An output file must be specified" << endl;
        ok = false;
    }
 
    if (format == NotGiven)
    {
        cerr << "An input format must be specified" << endl;
        ok = false;
    }
    
    if (ok && optind == argc)
    {
        cerr << "No input files specified" << endl;
        ok = false;
    }
    
    if (ok)
    {
        tagfile.open(taglist.c_str());
        if (!tagfile)
        {
            cerr << "Cannot open tag list file '" << taglist
                << "'" << endl;
            ok = false;
        }
    }

    if (ok && endname != "")
    {
        endfile.open(endname.c_str());
        if (!endfile)
        {
            cerr << "Cannot open ending rules file '" << endname
                << "'" << endl;
            ok = false;
        }
    }

    if (ok)
    {
        if (!textual)
        {
            outfile.open(outname.c_str(), ios::out | ios::binary);
        }
        else
        {
            outfile.open(outname.c_str());
        }

        if (!outfile)
        {
            cerr << "Cannot open output file '" << outname << "'" << endl;
            ok = false;
        }
    }

    if (!ok)
    {
        cerr << "Error(s) in command line" << endl;
        return -1;
    }
    else
    {
        TagSet        tagset;
        Trainer train(tagset);
        Lexicon       lex;
        char          tagscore_sep = '/';

        lex.setNumberTest(number);
        
        if (!train.start(tagfile, lex))
        {
            cerr << "Error in initialisation" << endl;
            return -1;
        }
        else
        {
            tagfile.close();
            
            int i;
            bool any = false;
            Endings rules;

            // Read endings file
            if (endfile.good())
            {
                bool ok = rules.readFile(endfile, tagset, tagscore_sep);
                endfile.close();
                
                if (!ok)
                {
                    cerr << "Error in initialisation" << endl;
                    return -1;
                }
            }

            TokenReader *token_reader
                = (format == Penn)   ? TokenReader::reader_penn(tagset)
                : (format == Tabbed) ? TokenReader::reader_tabbed(tagset)
                : NULL;

            for (i = optind ; i < argc ; i++)
            {
                ifstream infile;
                infile.open(argv[i]);
                if (!infile)
                {
                    cerr << "Cannot open input file '" << argv[i]
                        << "'" << endl;
                }
                else
                {
                    if (report)
                    {
                        cerr << argv[i] << endl;
                    }
                    any = true;
                    train.train(infile, echo, token_reader, casehold);
                    infile.close();
                }
            }

            delete token_reader;

            if (!any)
            {
                cerr << "No input files were valid" << endl;
                return -1;
            }
            else
            {
                train.end(corr, rules, outfile, reduce_level, textual);
                delete corr;
                outfile.close();
            }
        }
    }
    return 0;
}
